#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different classes to handle preferences
"""

from __future__ import print_function, division, absolute_import

import os
import json
import pickle
import logging

from tpDcc.libs.python import python, decorators

if python.is_python2():
    json_decode_error = Exception
else:
    json_decode_error = json.decoder.JSONDecodeError

logger = logging.getLogger('tpDcc-libs-python')

_broken_files = dict()


class PreferenceHandler(object):
    """
    Allows for a specific preference to be handled completely different than others
    Must reimplement set_preference_value and get_preference_value functions
    """

    @staticmethod
    @decorators.abstractmethod
    def get_preference_value():
        raise NotImplementedError

    @staticmethod
    @decorators.abstractmethod
    def set_preference_value():
        raise NotImplementedError


class PreferenceFile(dict):
    """
    Base Dictionary based preference file.
    """

    def __init__(self, path, handlers=None):

        self._path = path
        self._handlers = handlers or dict()
        if os.path.isfile(self._path):
            self.read()
        else:
            self.write()

        super(PreferenceFile, self).__init__()

    def __setitem__(self, key, value):
        self.read()
        if key in self._handlers:
            value = self._handlers.get(key).set_preference_value(value)
            if not value:
                value = '<external>'
        super(PreferenceFile, self).__setitem__(key, value)
        self.write()

    def __getitem__(self, key):
        self.read()
        if key in self._handlers:
            return self._handlers.get(key).get_preference_value()
        return super(PreferenceFile, self).__getitem__(key)

    @property
    def path(self):
        return self._path

    @property
    def handlers(self):
        return self._handlers

    @decorators.abstractmethod
    def write(self):
        """
        Writes local preferences to stored path
        Must be overwritten in custom preferences
        :return: str
        """

        raise NotImplementedError

    @decorators.abstractmethod
    def read(self):
        """
        Reads local preferences from stored path
        Must be overwritten in custom preferences
        :return: str
        """

        raise NotImplementedError

    def get(self, key, default=None):
        """
        Returns preference value. If not found, fallback to default
        :param key: str
        :param default: variant
        :return: variant
        """

        try:
            return self[key]
        except KeyError:
            return default

    def pop(self, key):
        """
        Overwrites base dict pop function to force to save of the preference file
        after removing the key
        :param key: str
        """

        super(PreferenceFile, self).pop(key)
        self.write()

    def set_handler(self, preference_key, handler):
        """
        Sets the handler to use for the given preference key
        :param preference_key: str
        :param handler: PrefHandler
        """

        self._handlers[preference_key] = handler

    def remove_handler(self, preference_key):
        """
        Removes any present handler from the given preference key
        :param preference_key: str
        """

        try:
            self._handlers.pop(preference_key)
        except KeyError:
            pass


class JsonPreference(PreferenceFile):

    def write(self):
        out = dict()
        out.update(self)
        with open(self._path, 'w+') as fh:
            json.dump(out, fh, indent=4, sort_keys=False, separators=(',', ': '))

    def read(self):
        contents = dict()
        if not os.path.isfile(self._path):
            return contents

        try:
            with open(self._path, 'r') as fh:
                contents = json.load(fh)
        except json_decode_error:
            _broken_files.setdefault(self._path, 0)
            times_hit = _broken_files[self._path]
            if times_hit < 3:
                logger.error('Invalid preference JSON file "{}"'.format(self._path))
            times_hit += 1
            _broken_files[self._path] = times_hit

        self.clear()
        self.update(contents)


class PicklePreference(PreferenceFile):

    def write(self):
        out = dict()
        out.update(self)
        with open(self._path, 'wb+') as fh:
            pickle.dump(out, fh, protocol=2)

    def read(self):
        contents = dict()
        if not os.path.isfile(self._path):
            return contents

        try:
            with open(self._path, 'r+b') as fh:
                contents = pickle.load(fh) if python.is_python2() else pickle.load(fh, encoding='bytes')
        except pickle.UnpicklingError:
            _broken_files.setdefault(self._path, 0)
            times_hit = _broken_files[self._path]
            if times_hit < 3:
                logger.error('Failed to load pickle preference "{}"'.format(self._path))
            times_hit += 1
            _broken_files[self._path] = times_hit

        self.clear()
        self.update(contents)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementations for Signal
"""

from __future__ import print_function, division, absolute_import


class Signal(object):
    """
    Simple single emission mechanism to allow for events withing functions to trigger mid-call callbacks
    """

    def __init__(self):
        self._callables = list()

    def connect(self, item):
        self._callables.append(item)

    def emit(self, *args, **kwargs):
        for item in self._callables:
            item(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utilities functions emails
"""

from __future__ import print_function, division, absolute_import

import email
import smtplib

try:
    from email.mime.text import MIMEText
except Exception:
    from email.MIMEText import MIMEText


class Email(object):
    def __init__(self, user=None, password=None):
        self._user = user
        self._password = password
        self._emails_list = list()
        self._server = None

        self._setup_message()

    def set_subject(self, subject):
        self._message['Subject'] = subject

    def add_message(self, text):
        self._message.attach(MIMEText(text))

    def send(self, list_of_emails):
        joined = email.Utils.COMMASPACE.join(list_of_emails)
        self._message['To'] = joined
        self._setup_server()
        self._server.sendmail(self._user, list_of_emails, self._message.as_string())
        self._server.quit()

    def _setup_server(self):
        self._server = smtplib.SMTP()
        self._server.set_debuglevel(1)

    def _setup_message(self):
        self._message = email.MIMEMultipart.MIMEMultipart()
        self._message['From'] = self._user
        self._message['Subject'] = ''
        self._message['To'] = list()


class LocalHost(Email, object):
    def __init__(self, user):
        super(LocalHost, self).__init__(user=user)

    def _setup_server(self):
        self._server = smtplib.SMTP('localhost')


class Gmail(Email, object):
    def _setup_server(self):
        super(Gmail, self)._setup_server()

        smtp_host = 'smtp.gmail.com'
        smtp_port = 587
        self._server.connect(smtp_host, smtp_port)
        self._server.ehlo()
        self._server.starttls()
        self._server.login(self._user, self._password)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility methods related to write/read (IO) text files
"""


from __future__ import print_function, division, absolute_import


import os
import sys
import stat
import json
import string
import shutil
import filecmp
import logging
import getpass
import datetime
import traceback
import subprocess
from tempfile import mkstemp
from shutil import move

from tpDcc.libs.python import python

LOGGER = logging.getLogger('tpDcc-libs-python')


class FileManager(object):
    """
    Base class to deal with file writing and reading
    """

    def __init__(self, file_path, skip_warning=False):
        """
        :param file_path: str, path to the file to work with
        :param skip_warning: bool, Whether to print warnings or not
        """

        self.file_path = file_path
        self.open_file = None

        if not skip_warning:
            self.check_path(warning_text='Path {} is invalid!'.format(file_path))

    def get_open_file(self):
        """
        Returns managed file and opens it
        :return:  str
        """

        return self.open_file()

    def read_file(self):
        """
        Opens managed file to read
        """

        self.check_file(warning_text='File {} is invalid!'.format(self.file_path))
        self.open_file = open(self.file_path, 'r')

    def write_file(self):
        """
        Opens managed file to write data into (removing any previous data)
        """

        # self.check_file(warning_text='File {} is invalid!'.format(self.file_path))
        self.open_file = open(self.file_path, 'w')

    def append_file(self):
        """
        Opens managed file to append data to the current file data
        """

        self.check_file(warning_text='File {} is invalid!'.format(self.file_path))
        self.open_file = open(self.file_path, 'a')

    def close_file(self):
        """
        Close managed file
        """

        if self.open_file:
            self.open_file.close()

    def check_folder(self, warning_text=None):
        """
        Check if a folder is an invalid one and raise error if necessary
        :param warning_text: str
        :return: bool
        """

        from tpDcc.libs.python import path

        if not path.is_dir(self.file_path):
            if warning_text is not None:
                raise NameError(str(warning_text))
            return False

        return True

    def check_path(self, warning_text=None):
        """
        Check if a path to a file is invalid and raise error if necessary
        :param warning_text: str
        :return: bool
        """

        from tpDcc.libs.python import path

        dir_name = path.get_dirname(self.file_path)

        if not path.is_dir(dir_name):
            if warning_text is not None:
                raise UserWarning(str(warning_text))
            return False

        return True

    def check_file(self, warning_text=None):
        """
        Check if a file is an invalid one and raise error if necessary
        :param warning_text: str
        :return: bool
        """

        from tpDcc.libs.python import path

        if not path.is_file(self.file_path):
            if warning_text is not None:
                raise UserWarning(str(warning_text))
            return False

        return True


class FileReader(FileManager, object):
    """
    Class to deal with file read operations
    """

    def __init__(self, file_path):
        super(FileReader, self).__init__(file_path=file_path)

    def read(self):
        """
        Read managed file
        :return: list<str>, list of file lines
        """

        self.read_file()
        lines = self._get_lines()
        self.close_file()

        return lines

    def _get_lines(self):
        try:
            lines = self.open_file.read()
        except Exception:
            return []

        return get_text_lines(lines)


class FileWriter(FileManager, object):
    """
    Class to deal with file write operations
    """

    def __init__(self, file_path):
        super(FileWriter, self).__init__(file_path=file_path)

        from tpDcc.libs.python import osplatform

        osplatform.get_permission(file_path)
        self.append = False

    def write_file(self):
        """
        Overrides write file. This function creates the file if it does not exists
        If append is True, then append any lines to the file instead or replacing
        """

        if self.append:
            self.append_file()
        else:
            super(FileWriter, self).write_file()

    def set_append(self, append):
        """
        If True, next write operations will append new lines to end of documents otherwise the text
        content will be replaced entirely
        :param append: bool
        """

        self.append = append

    def write_line(self, line):
        """
        Writes a single line to the file and closes managed file after write operation
        :param line: str, line to add to the file
        """

        self.write_file()
        try:
            self.open_file.write('%s\n' % line)
        except Exception:
            pass
        self.close_file()

    def write_json(self, dict_data):
        """
        Writes given JSON data (dict) into file
        :param dict_data: dict
        """

        self.write_file()
        try:
            if python.is_python2():
                json.dump(dict_data, self.open_file, indent=4, sort_keys=False)
            else:
                json.dump(dict(dict_data), self.open_file, indent=4, sort_keys=False)
        except Exception as exc:
            LOGGER.exception('Impossible to save JSON file: "{}"'.format(exc))
        self.close_file()

    def write(self, lines, last_line_empty=True):
        """
        Write the given list of lines to the managed file
        :param lines: list<str>, list of lines to write to the managed file
        :param last_line_empty: bool, whether or not to add a line after the last line
        """

        self.write_file()

        try:
            inc = 0
            for line in lines:
                if inc == len(lines) - 1 and not last_line_empty:
                    self.open_file.write(str('%s' % line))
                    break
                self.open_file.write(str('%s\n' % line))
                inc += 1
        except Exception:
            print('Could not write to file {}'.format(self.file_path))

        self.close_file()


class FileVersion(object):
    """
    Utility class to version files or folders
    """

    def __init__(self, file_path):

        from tpDcc.libs.python import path

        self.file_path = file_path
        if file_path:
            self.filename = path.get_basename(directory=self.file_path)
            self._path = path.get_dirname(file_path)
            self._version_folder_name = '__version__'
            self._version_name = 'version'
            self._version_folder = None
            self.comment_file = None
            self.updated_old = False

    def get_version_name(self):
        return self._version_name

    def set_version_name(self, version_name):
        self._version_name = version_name

    def get_version_folder_name(self):
        return self._version_folder_name

    def set_version_folder_name(self, version_folder_name):
        self._version_folder_name = version_folder_name

    def get_version_folder(self):
        return self._version_folder

    def set_version_folder(self, version_folder):
        self._version_folder = version_folder

    version_name = property(get_version_name, set_version_name)
    version_folder_name = property(get_version_folder_name, set_version_folder_name)
    version_folder = property(get_version_folder, set_version_folder)

    def has_versions(self):
        """
        Returns whether the version file already has versions folder created or not
        :return: bool
        """

        from tpDcc.libs.python import path

        version_folder = self._get_version_folder()
        if path.is_dir(version_folder):
            return True

    def get_latest_version(self):
        """
        Returns the file path to the latest version
        :return: str
        """

        from tpDcc.libs.python import path

        versions = self.get_versions()
        latest_version = versions[-1]

        return path.join_path(self.file_path, '{0}/{1}'.format(self.version_folder_name, latest_version))

    def get_versions(self, return_version_numbers_also=False):
        """
        Get file paths of all version
        :param return_version_numbers_also: Whether the number of the versions should be returned also
        :return: list
        """

        from tpDcc.libs.python import folder, sort

        version_folder = self._get_version_folder()
        files = folder.get_files_and_folders(directory=version_folder)
        if not files:
            return None

        number_list = list()
        pass_files = list()

        for file_path in files:
            if not file_path.startswith(self.version_name):
                continue
            split_name = file_path.split('.')
            if not len(split_name) == 2:
                continue

            number = int(split_name[1])
            number_list.append(number)
            pass_files.append(file_path)

        if not pass_files:
            return

        quick_sort = sort.QuickNumbersListSort(list_of_numbers=number_list)
        quick_sort.set_follower_list(pass_files)
        pass_files = quick_sort.run()

        pass_dict = dict()
        for i in range(len(number_list)):
            pass_dict[pass_files[0][i]] = pass_files[1][i]

        if not return_version_numbers_also:
            return pass_dict
        else:
            return pass_dict, pass_files[0]

    def get_version_numbers(self):
        """
        Return file version numbers of all versions
        :return: list<int>, list of version numbers
        """

        from tpDcc.libs.python import folder

        version_folder = self._get_version_folder()
        files = folder.get_files_and_folders(directory=version_folder)
        if not files:
            return

        number_list = list()
        for file_path in files:
            if not file_path.startswith(self.version_name):
                continue
            split_name = file_path.split('.')
            if not len(split_name) == 2:
                continue
            num = int(split_name[1])
            number_list.append(num)

        return number_list

    def save_comment(self, comment=None, version_file=None):
        """
        Saves a comment to the version file
        :param comment: str, commend to add to the version file
        :param version_file: str, version file
        """

        version = version_file.split('.')
        if version:
            version = version[-1]

        user = getpass.getuser()

        if not comment:
            comment = '-'
        comment.replace('"', '\"')

        comment_file = FileWriter(file_path=self.comment_file)
        comment_file.set_append(True)
        comment_file.write(['version = {0}; comment = "{1}"; user = "{2}"'.format(version, comment, user)])
        comment_file.close_file()

    def save(self, comment=None):
        """
        Saves a new version file
        :param comment: str
        :return: str, new version file name
        """

        from tpDcc.libs.python import folder, path

        if not comment:
            comment = '-'
        comment = comment.replace('\n', '   ').replace('\r', '   ')

        self._create_version_folder()
        self._create_comment_file()

        unique_file_name = self._increment_version_file_name()

        if path.is_dir(self.file_path):
            folder.copy_folder(directory=self.file_path, directory_destination=unique_file_name)
        elif path.is_file(self.file_path):
            copy_file(file_path=self.file_path, file_path_destination=unique_file_name)

        self.save_comment(comment=comment, version_file=unique_file_name)

        return unique_file_name

    def get_version_data(self, version_number):
        """
        Returns the version data (comment and user) of the given version number
        :param version_number: int, version number
        :return: list<str, str>, tuple with comment and user of the given version
        """

        from tpDcc.libs.python import path

        file_path = self._get_comment_path()
        if not file_path:
            return None, None

        if path.is_file(file_path):
            read = FileReader(file_path=file_path)
            lines = read.read()

            version = None
            comment = None
            user = None

            for line in lines:
                start_index = line.find('"')
                if start_index > -1:
                    end_index = line.find(';')
                    sub_part = line[start_index + 1:end_index]
                    sub_part = sub_part.replace('"', '\\"')
                    line = line[:start_index + 1] + sub_part + line[end_index:]

                try:
                    exec(line)
                except Exception:
                    pass

                if version == version_number:
                    return comment, user

        return None, None

    def get_all_versions_data(self):
        """
        Returns all the version data (comment, user, file_size, modified and version_file) of all the versions
        :return: list<str, str, str, str, str, str>, tuple version, comment, user, file_size, modified, file_version
        """

        from tpDcc.libs.python import path

        versions = self.get_versions(return_version_numbers_also=True)
        if not versions:
            return
        else:
            version_paths = versions[0]
            version_numbers = versions[1]

        file_path = self._get_comment_path()
        if not file_path:
            return []

        datas = list()
        if path.is_file(file_path):
            read = FileReader(file_path)
            lines = read.read()
            for line in lines:
                line_info_dict = dict()
                version = None
                comment = None
                user = None
                file_size = None
                modified = None

                split_line = line.split(';')
                for sub_line in split_line:
                    assigment = sub_line.split('=')
                    if assigment and assigment[0]:
                        name = assigment[0].strip()
                        value = assigment[1].strip()
                        line_info_dict[name] = value

                # Version
                if 'version' not in line_info_dict:
                    continue
                version = int(line_info_dict['version'])
                if version not in version_numbers:
                    continue

                # Comment
                if 'comment' in line_info_dict:
                    comment = line_info_dict['comment']
                    comment = comment[1:-1]

                # User
                if 'user' in line_info_dict:
                    user = line_info_dict['user']
                    user = user[1:-1]

                # Version File
                version_file = version_paths[(version)]
                version_file = path.join_path(self.file_path, '{0}/{1}'.format(self.version_folder_name, version_file))

                # File Size
                file_size = get_file_size(file_path=version_file)

                # Modified
                modified = get_last_modified_date(file_path=version_file)

                datas.append([version, comment, user, file_size, modified, version_file])

        return datas

    def get_version_path(self, version_number):
        """
        Returns the path to the given version number
        :param version_number: int, version number
        :return: str, path to the version
        """

        return self._get_version_path(version_number=version_number)

    def get_version_comment(self, version_number):
        """
        Returns the comment of the given version number
        :param version_number: int, version number
        :return:  str, version_number comment
        """

        comment, user = self.get_version_data(version_number)
        return comment

    def _get_version_folder(self):
        from tpDcc.libs.python import path

        if path.is_file(self.file_path):
            dir_name = path.get_dirname(self.file_path)
            version_path = path.join_path(dir_name, self._version_folder_name)
        else:
            version_path = path.join_path(self.file_path, self._version_folder_name)

        return version_path

    def _get_version_path(self, version_number):
        from tpDcc.libs.python import path
        return path.join_path(self._get_version_folder(), self._version_name + '.' + str(version_number))

    def _get_version_number(self, file_path):
        from tpDcc.libs.python import name
        version_number = name.get_last_number(input_string=file_path)
        return version_number

    def _get_comment_path(self):
        from tpDcc.libs.python import path
        version_folder = self._get_version_folder()
        file_path = None
        if version_folder:
            file_path = path.join_path(version_folder, 'comments.txt')

        return file_path

    def _create_version_folder(self):
        from tpDcc.libs.python import folder
        self._version_folder = folder.create_folder(name=self._version_folder_name, directory=self._path)

    def _create_comment_file(self):
        self.comment_file = create_file(filename='comments.txt', directory=self._version_folder)

    def _increment_version_file_name(self):
        from tpDcc.libs.python import path
        version_path = path.join_path(self._version_folder, self._version_name + '.1')
        return path.unique_path_name(directory=version_path)


def open_browser(file_path):
    """
    Open the file browser to the path specified
    :param file_path: str, filename with path
    :return:
    """

    from tpDcc.libs.python import osplatform, path

    if not path.is_file(file_path) and not path.is_dir(file_path):
        return

    if osplatform.is_windows():
        os.startfile(file_path)
    elif osplatform.is_linux():
        try:
            os.system('gio open {}'.format(file_path))
        except Exception:
            try:
                opener = 'open' if sys.platform == 'darwin' else 'xdg-open'
                subprocess.call([opener, file_path])
            except Exception as e:
                os.system('gnome-terminal --working-directory={}'.format(file_path))


def create_file(filename, directory=None, make_unique=False):
    """
    Creates a file
    :param filename: str, name of the new file
    :param directory: str, directory of the new file
    :param make_unique: bool, whether to make the name unique or not
    :return: variant, str || bool, filename with path or False if create file failed
    """

    from tpDcc.libs.python import name, path, osplatform

    if directory is None:
        directory = path.get_dirname(filename)
        filename = path.get_basename(filename)

    filename = name.clean_file_string(filename)
    full_path = path.join_path(directory, filename)

    if make_unique:
        full_path = path.unique_path_name(full_path)

    open_file = None
    try:
        open_file = open(full_path, 'a')
        open_file.close()
    except Exception:
        if open_file:
            open_file.close()
        return False

    osplatform.get_permission(full_path)

    return full_path


def copy_file(file_path, file_path_destination):
    """
    Copies the given file to a new given directory
    :param file_path: str, file to copy with full path
    :param file_path_destination: str, destination directory where we want to copy the file into
    :return: str, the new copied path
    """

    from tpDcc.libs.python import path, osplatform

    osplatform.get_permission(file_path)

    if path.is_file(file_path):
        if path.is_dir(file_path_destination):
            file_name = path.get_basename(file_path)
            file_path_destination = path.join_path(file_path_destination, file_name)
        shutil.copy2(file_path, file_path_destination)

    return file_path_destination


def move_file(path1, path2):
    """
    Moves the file pointed by path1 under the directory path2
    :param path1: str, file with full path
    :param path2: str, path where path1 should be move into
    :return: bool, Whether the move operation was successfully
    """

    try:
        shutil.move(path1, path2)
    except Exception:
        print('Failed to move {0} to {1}'.format(path1, path2))
        return False

    return True


def delete_file(name, directory=None, show_warning=True):
    """
    Delete the file by name in the directory
    :param name: str, name of the file to delete
    :param directory: str, the directory where the file is stored
    :param show_warning: bool
    :return: str, file path that was deleted
    """

    from tpDcc.libs.python import path, osplatform

    if not directory:
        full_path = name
    else:
        full_path = path.join_path(directory, name)
    if not path.is_file(full_path):
        if show_warning:
            print('File "{}" was not deleted.'.format(full_path))
        return full_path

    try:
        osplatform.get_permission(full_path)
    except Exception:
        pass
    try:
        os.remove(full_path)
    except Exception:
        pass

    return full_path


def rename_file(name, directory, new_name, new_version=False):
    """
    Renames the give nfile in the directory with a new name
    :param name:
    :param directory:
    :param new_name:
    :param new_version:
    :return:
    """

    from tpDcc.libs.python import path

    full_path = path.join_path(directory, name)
    if not path.is_file(full_path):
        return full_path

    new_full_path = path.join_path(directory, new_name)
    if path.is_file(new_full_path):
        print('A file named {} already exists in the directory: {}'.format(new_name, directory))
        return full_path

    os.chmod(full_path, 0o777)
    os.rename(full_path, new_full_path)

    return new_full_path


def write_to_file(file_name, text_to_write):
    """
    Open a file and overwrite its contents of that file with new content
    """

    if os.path.exists(file_name):
        read_only_or_writeable = os.stat(file_name)[0]
        if read_only_or_writeable != stat.S_IWRITE:
            os.chmod(file_name, stat.S_IWRITE)
    with open(file_name, 'w') as file:
        file.write(text_to_write)


def append_to_file(file_name, text_to_add):
    """
    Open a file and add new context to its existing text
    """

    if os.path.exists(file_name):
        readOnlyOrWriteable = os.stat(file_name)[0]
        if readOnlyOrWriteable != stat.S_IWRITE:
            os.chmod(file_name, stat.S_IWRITE)
    with open(file_name, 'a') as file:
        file.write(text_to_add)


def replace(file_path, pattern, subst):
    """
    Replaces one string from another string in a given file
    :param file_path: str, path to the file
    :param pattern: search to be replaced
    :param subst: string that will replace the old one
    """

    # Create temp file
    fh, abs_path = mkstemp()
    with os.fdopen(fh, 'w') as new_file:
        with open(file_path) as old_file:
            for line in old_file:
                # TODO: This gives when working with non ascii codecs, solve ethis without try/except
                try:
                    new_file.write(line.replace(pattern, subst))
                except Exception:
                    pass
    # Remove original file
    os.remove(file_path)

    # Move new file
    move(abs_path, file_path)


def remove_extension(file_path):
    """
    Removes extension of the given file path
    For example, C:/test/rig.py will return rig
    :param file_path: str
    :return: str
    """

    split_path = file_path.split('.')
    new_name = file_path
    if len(split_path) > 1:
        new_name = '.'.join(split_path[:-1])

    return new_name


def is_newer(file1, file2):
    """
    Returns true if file1 is newer than file2.
    :param str file1: first file to compare.
    :param str file2: second file to compare.
    :return: bool, True if file1 is newer or False otherwise
    """

    if not os.path.exists(file1) or not os.path.exists(file2):
        return False

    time1 = os.path.getmtime(file1)
    time2 = os.path.getmtime(file2)
    return time1 > time2


def is_file_in_dir(filename, directory):
    """
    Returns True if the given file name is contained in the given directory
    :param filename: str, file name
    :param directory: str, directory name including path
    :return: bool
    """

    from tpDcc.libs.python import path

    file_path = path.join_path(directory, filename)
    return os.path.isfile(file_path)


def is_file_empty(file_path):
    """
    Returns whether or not given file is empty
    :param file_path: str
    :return: bool, True if file is empty; False otherwise.
    """

    if not os.path.isfile(file_path):
        return True

    return os.stat(file_path).st_size == 0


def is_same_date(file1, file2):
    """
    Returns True if the given files have the same date
    :param file1: str, file name including path
    :param file2: str, file name including path
    :return: bool
    """

    if not file1 and file2 or file1 and not file2:
        return False
    if not file1 and not file2:
        return True

    date1 = os.path.getmtime(file1)
    date2 = os.path.getmtime(file2)
    if not date1 and not date2:
        return True

    if date1 and date2:
        value = date1 - date2
        if abs(value) < 0.01:
            return True

    return False


def get_file_text(file_path):
    """
    Get the text stored in a file in a unique string (without parsing)
    :param file_path: str
    :return: str
    """

    try:
        with open(file_path, 'r') as open_file:
            lines = open_file.read()
    except Exception:
        return list()

    return lines


def get_file_size(file_path, round_value=2):
    """
    Returns the size of the given file
    :param file_path: str
    :param round_value: int, value to round size to
    :return: str
    """

    size = os.path.getsize(file_path)
    size_format = round(size * 0.000001, round_value)

    return size_format


def get_size(file_path, round_value=2):
    """
    Return the size of the given directory or file path
    :param file_path: str
    :param round_value: int, value to round size to
    :return: int
    """

    from tpDcc.libs.python import folder, path

    size = 0
    if path.is_dir(file_path):
        size = folder.get_folder_size(file_path, round_value)
    if path.is_file(file_path):
        size = get_file_size(file_path, round_value)

    return size


def get_last_modified_date(file_path, reverse_date=False):
    """
    Returns the last date given file was modified
    :param file_path: str
    :param reverse_date: bool
    :return: str, formatted date and time
    """

    mtime = os.path.getatime(file_path)

    date_value = datetime.datetime.fromtimestamp(mtime)
    year = date_value.year
    month = date_value.month
    day = date_value.day

    hour = str(date_value.hour)
    minute = str(date_value.minute)
    second = str(int(date_value.second))

    if len(hour) == 1:
        hour = '0' + hour
    if len(minute) == 1:
        minute = '0' + minute
    if len(second) == 1:
        second = second + '0'

    if reverse_date:
        return '{0}-{1}-{2}  {3}:{4}:{5}'.format(year, month, day, hour, minute, second)
    else:
        return '{0}-{1}-{2}  {3}:{4}:{5}'.format(day, month, year, hour, minute, second)


def get_file_date(file_path):
    """
    Returns date the given file was created
    :param file_path: str
    :return: str
    """

    date_file = 0
    if os.path.isfile(file_path):
        st_file = os.stat(file_path)
        date_file = st_file[stat.ST_MTIME]

    return date_file


def copy_file_date(original_file_path, target_file_path):
    """
    Copies the creation date of one file to another one
    :param original_file_path: str
    :param target_file_path: str
    """

    if os.path.isfile(original_file_path) and os.path.isfile(target_file_path):
        date_file = get_file_date(original_file_path)
        os.utime(target_file_path, (date_file, date_file))


def get_file_lines(file_path):
    """
    Get the text lines from a file
    :param file_path: str, file name of the text to read
    :return: str
    """

    text = get_file_text(file_path)
    if not text:
        return list()

    return get_text_lines(text)


def get_text_lines(text):
    """
    Get all lines from a text storing each lines as a different item in a list
    :param text: str, text to get lines from
    :return: list<str>
    """

    text = text.replace('\r', '')
    lines = text.split('\n')

    return lines


def write_replace(file_path, data_to_write):
    """
    Writes given data into given file path (replacing already existing content)
    :param file_path: str
    :param data_to_write:
    """

    with open(file_path, 'w') as file_handle:
        try:
            file_handle.write(data_to_write)
        except Exception:
            LOGGER.warning('Could not write: {} | {}'.format(data_to_write, traceback.format_exc()))


def write_lines(file_path, lines, append=False):
    """
    Writes a list of text lines to a file. Every entry in the list is a new line
    :param file_path: str, filename and path
    :param lines: list<str>, list of text lines in which each entry is a new line
    :param append: bool, Whether to append the text or replace it
    """

    from tpDcc.libs.python import python, osplatform

    permission = osplatform.get_permission(file_path)
    if not permission:
        return

    write_string = 'a' if append else 'w'

    lines = python.force_list(lines)
    text = '\n'.join(map(str, lines))
    if append:
        text = '\n' + text

    with open(file_path, write_string) as open_file:
        open_file.write(text)


def is_same_text_content(file1, file2):
    """
    Returns True if the given text files contains the same text or False otherwise
    :param file1: str, file path to the first text
    :param file2: str, file path to the second text
    :return: bool
    """

    return filecmp.cmp(file1, file2)


def get_files(root_directory, filter_text=''):
    """
    Returns files found in the given directory
    :param root_directory: str
    :param filter_text: str
    :return: list(str)
    """

    found = list()

    files = os.listdir(root_directory)
    for filename in files:
        if filter_text and filename.find(filter_text) == -1:
            continue
        file_path = os.path.join(root_directory, filename)
        if os.path.isfile(file_path):
            found.append(filename)

    return found


def file_has_info(file_path):
    """
    Check if the given file size is bigger than 1.0 byte
    :param file_path: str, ful file path of the file to check
    :return: bool
    """

    file_stats = os.stat(file_path)
    if file_stats.st_size < 1:
        return False

    return True


def get_lock_name(file_path):
    """
    Returns lock file of the given file
    :param file_path: str
    :return: str
    """

    return '{}.lock'.format(file_path)


def is_locked(file_path):
    """
    Returns whether or not given file is locked
    :param file_path: str
    :return: bool
    """

    return os.path.isfile(get_lock_name(file_path))


def lock(file_path):
    """
    Creates lock file of the given file
    :param file_path: str
    :return: str
    """

    lock_file = get_lock_name(file_path)
    create_file(lock_file)

    return lock_file


def remove_lock(file_path):
    """
    Removes lock file of the given file
    :param file_path: str
    :return: bool
    """

    lock_file = get_lock_name(file_path)
    if not os.path.isfile(lock_file):
        return False

    return delete_file(lock_file)


def get_latest_file(file_paths, only_return_one_match=True):
    """
    Returns the latest created file from a list of file paths
    :param file_paths: list(str)
    :param only_return_one_match: bool
    :return: list(str) or str
    """
    last_time = 0
    times = dict()

    for file_path in file_paths:
        mtime = os.stat(file_path).st_mtime
        if mtime not in times:
            times[mtime] = list()
        times[mtime].append(file_path)
        if mtime > last_time:
            last_time = mtime

    if not times:
        return

    if only_return_one_match:
        return times[mtime][0]
    else:
        return times[mtime]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Implementation of Dijkstra short distance algorithm
https://gist.github.com/econchick/4666413
"""

from __future__ import print_function, division, absolute_import

from collections import defaultdict, deque


class Graph(object):
    """
    Dijkstra closest path (for NURBS and lattice)
    https://gist.github.com/econchick/4666413
    """

    def __init__(self):
        self.nodes = set()
        self.edges = defaultdict(list)
        self.distances = {}

    def add_node(self, value):
        self.nodes.add(value)

    def add_edge(self, from_node, to_node, distance):
        self.edges[from_node].append(to_node)
        self.edges[to_node].append(from_node)
        self.distances[(from_node, to_node)] = distance


def dijkstra(graph, initial):
    visited = {initial: 0}
    path = {}

    nodes = set(graph.nodes)

    while nodes:
        min_node = None
        for node in nodes:
            if node in visited:
                if min_node is None:
                    min_node = node
                elif visited[node] < visited[min_node]:
                    min_node = node
        if min_node is None:
            break

        nodes.remove(min_node)
        current_weight = visited[min_node]

        for edge in graph.edges[min_node]:
            try:
                weight = current_weight + graph.distances[(min_node, edge)]
            except Exception:
                continue
            if edge not in visited or weight < visited[edge]:
                visited[edge] = weight
                path[edge] = min_node

    return visited, path


def shortest_path(graph, origin, destination):
    visited, paths = dijkstra(graph, origin)
    full_path = deque()
    _destination = paths[destination]

    while _destination != origin:
        full_path.appendleft(_destination)
        _destination = paths[_destination]

    full_path.appendleft(origin)
    full_path.append(destination)

    return visited[destination], list(full_path)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different classes to handle settings
"""

from __future__ import print_function, division, absolute_import

import os
import ast
import time
import logging
from collections import OrderedDict
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

from tpDcc.libs.python import python, fileio, path, jsonio

LOGGER = logging.getLogger('tpDcc-libs-python')


class FileSettings(object):
    def __init__(self):
        self.directory = None
        self.file_path = None

        self.settings_dict = OrderedDict()
        self.write = None

    def data(self):
        """
        Return data dictonary contained in the settings
        :return: dict
        """

        return self.settings_dict

    def get(self, name, default=None):
        """
        Get a stored setting
        :param name: str, name of the setting to retrieve. Returns None, if setting is not found
        :param name: default, variant, default value to return if setting is not found
        :return: variant, None || str
        """

        # Force settings reload
        self.reload()

        if name in self.settings_dict:
            return self.settings_dict[name]

        if default is not None:
            return default

    def set(self, name, value):
        """
        Set the value of a specific setting. If the setting does not exists, the setting is created with the
        given value
        :param name: str, name of the setting
        :param value: varinat, value of the setting
        """

        self.settings_dict[name] = value
        self._write()

    def get_settings(self):
        """
        Returns a list with all the settings stored
        :return: list<list<str, variant>>
        """

        found = list()
        for setting in self.settings_dict.keys():
            found.append([setting, self.settings_dict[setting]])

        return found

    def get_file(self):
        """
        Retuns the file path of the settings file
        :return: str
        """

        return self.file_path

    def set_directory(self, directory, filename='settings.cfg'):
        """
        Set the file that is used to stored settins on file
        :param directory: str
        :param filename: str
        :return: str
        """

        self.directory = directory
        self.file_path = fileio.create_file(filename=filename, directory=directory)
        self._read()

        return self.file_path

    def has_setting(self, name):
        """
        Returns if a specific name is stored
        :param name: str, name of the setting
        :return: bool
        """

        return name in self.settings_dict

    def has_settings(self):
        """
        Returns if there are settings stored or not
        :return: bool
        """

        if self.settings_dict:
            return True
        return False

    def reload(self):
        """
        Forces the reading of the settings
        """

        self._read()

    def clear(self):
        """
        Cleans the stored settings
        """

        self.settings_dict = OrderedDict()
        self._write()
    # endregion

    # region Private Functions
    def _read(self):
        """
        Internal function used to read settings from file
        """
        if not self.file_path:
            return

        lines = fileio.get_file_lines(self.file_path)
        if not lines:
            return

        self.settings_dict = OrderedDict()

        for line in lines:
            if not line:
                continue
            split_line = line.split('=')
            name = split_line[0].strip()
            value = split_line[-1]
            if not value:
                continue
            value = path.clean_path(value)
            try:
                value = eval(str(value))
            except Exception:
                value = str(value)

            self.settings_dict[name] = value

    def _write(self):
        """
        Internal function that writes stored settings into text file
        """

        lines = list()
        for key in self.settings_dict.keys():
            value = self.settings_dict[key]
            if python.is_string(value):
                value = '"{}"'.format(value)
            if value is None:
                value = 'None'
            line = '{0} = {1}'.format(key, str(value))
            lines.append(line)

        write = fileio.FileWriter(file_path=self.file_path)
        try:
            write.write(lines)
        except Exception:
            LOGGER.debug('Impossible to write in {}'.format(self.file_path))
            time.sleep(.1)
            write.write(lines)


class JSONSettings(FileSettings, object):
    def __init__(self, directory=None, filename='settings.json'):
        super(JSONSettings, self).__init__()
        if directory:
            self.set_directory(directory, filename if filename and filename.endswith('.json') else 'settings.json')

    def set_directory(self, directory, filename='settings.json'):
        self.directory = directory

        # Check that given file name is a valid JSON file
        if not filename.endswith('.json'):
            old = path.join_path(directory, filename)
            if path.is_file(old):
                self.file_path = old
                self._read()
                return

        self.file_path = fileio.create_file(filename=filename, directory=directory)

        self._read()

        return self.file_path

    def _write(self):
        """
        Override function to add support to write JSON files
        """

        file_path = self._get_json_file()
        if not file_path:
            return

        writer = fileio.FileWriter(file_path)
        writer.write_json(list(self.settings_dict.items()))

    def _read(self):
        """
        Override function to add support to read JSON files
        """

        if not self._has_json_file():
            self.settings_dict = OrderedDict()
            return

        file_path = self._get_json_file()
        if not file_path:
            return
        self.file_path = file_path

        try:
            data = OrderedDict(jsonio.read_file(file_path))
        except Exception:
            self.settings_dict = OrderedDict()
            return

        self.settings_dict = data

    def _get_json_file(self):
        """
        Internal function that returns JSON file where settings are stored
        :return: str
        """

        if not self.file_path:
            return

        settings_directory = path.get_dirname(self.file_path)
        name = path.get_basename(self.file_path, with_extension=False)
        file_path = fileio.create_file(name + '.json', settings_directory)
        if not file_path:
            test_path = path.join_path(settings_directory, name + '.json')
            if path.is_file(test_path):
                file_path = test_path

        return file_path

    def _has_json_file(self):
        """
        Checks if the JSON file where settings should be stored exists or not
        :return: bool
        """

        if not self.file_path:
            return False

        settings_directory = path.get_dirname(self.file_path)
        name = path.get_basename(self.file_path, with_extension=False)
        file_path = path.join_path(settings_directory, name + '.json')
        if path.is_file(file_path):
            return True

        return False


class INISettings(object):
    def __init__(self, filename):
        """
        Constructor
        :param filename: str, INI filename
        """

        self._file = filename
        self._parser = configparser.RawConfigParser()
        self._is_dirty = False
        try:
            self._parser.read(self._file)
        except Exception:
            LOGGER.warning('Impossible to read INI config file "{}"'.format(filename))
        self._section = list()

    def __enter__(self):
        """
        Utility function to be used with Python 'with' functino
        """

        return self

    def __exit__(self, type_, value, tb):
        """
        Utility function bo be used with Python 'with' function. Release the raw file and parser
        """

        self.close()
        return False

    def get_section(self):
        """
        Returns the current stack's section
        """

        return self._section[len(self._section) - 1].upper() if self._section else None

    section = property(get_section)

    def save(self):
        """
        Writes INI file to disk
        """

        if self._is_dirty:
            if not os.path.exists(os.path.dirname(self._file)):
                os.makedirs(os.path.dirname(self._file))

            with open(self._file, 'w') as f:
                self._parser.write(f)

        self._is_dirty = False

    def show_in_explorer(self):
        """
        Displays the INI file explorer
        """

        import subprocess
        subprocess.Popen(r'explorer /select, {0}'.format(self._file))

    def get(self, section, option, default, eval_=False):
        """
        Returns the value associated with a key in a section
        :param section: str, section containing the key
        :param option: str, key to use in retrieval
        :param default: object, value to use if the key is not defined
        :param eval_: bool, if True, will attempt to evaluate the string returned into a Python type
        :return: variant, associated value if successful; value passed in as default otherwise
        """

        if self._parser.has_section(section) and self._parser.has_option(section, option):
            if isinstance(default, bool):
                return self._parser.getboolean(section, option)
            if isinstance(default, float):
                return self._parser.getfloat(section, option)
            int_types = (int, long) if python.is_python2() else (int)
            if isinstance(default, int_types):
                return self._parser.getint(section, option)
            if eval_:
                return ast.literal_eval(self._parser.get(section, option))
            return self._parser.get(section, option)
        return default

    def set(self, section, option, value):
        """
        Sets the value with an associated key in the respective section
        :param section: str, section containing the key
        :param option: str, key to use in storing the value
        :param value: object, value to store
        """

        if section:
            if not self._parser.has_section(section):
                self._parser.add_section(section)
            self._parser.set(section, option, str(value))
            self._is_dirty = True

    def remove(self, section, option):
        """
        Removes a key in a section
        :param section: str, section containing the key
        :param option: str, key to remove
        :return: bool, True if successful or False, otherwise
        """

        success = False
        if section:
            if self._parser.has_section(section):
                success = self._parser.remove_option(section, option)

        if success:
            self._is_dirty = True

        return success

    def push_section(self, section):
        """
        Pushes a new section onto the section stack
        :param section: str, INI section to push
        :return: str
        """

        self._section.append(section)

    def pop_section(self):
        """
        Removes the current section from the section stack
        :return: str
        """

        self._section.pop()

    def import_option(self, option, default, eval_=False):
        """
        Returns the value associated with a key in the section at the top of the section stack
        :param option: str, key to use in retrieval
        :param default: str, value to use if the key is not defined
        :param eval_: bool, if True, will attempt to evaluate the string returned into a Python type
        :return: associated value if successful or the value passed in as default otherwise
        """

        return self.get(self.section, option, default, eval_)

    def delete_option(self, option):
        """
        Removes the associated key in the section at the top of the section stack
        :param option: key to use in storing the value
        """

        return self.remove(self.section, option)

    def export_option(self, option, value):
        """
        Sets the value with an associated key in the section at the top of the stack's section
        :param option: str, key to use in storing the value
        :param value: value to store
        """

        self.set(self.section, option, value)

    def close(self):
        """
        Closes INI file
        """

        self.save()
        self._parser = None



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains bezier related functions
"""

from __future__ import print_function, division, absolute_import

import math


def binomial(i, n):
    return math.factorial(n) / float(math.factorial(i) * math.factorial(n - i))


def bernstein(t, i, n):
    return binomial(i, n) * (t ** i) * ((1 - t) ** (n - i))


def bezier(t, points):
    n = len(points) - 1
    x = 0
    y = 0
    for i, pos in enumerate(points):
        bern = bernstein(t, i, n)
        x += pos[0] * bern
        y += pos[1] * bern

    return x, y


def bezier_curve_y_from_x(index_x, points):

    def take_closest(num, collection):
        return min(collection, key=lambda x_num: abs(x_num - num))

    full_list = dict()
    for i in range(101):
        x, y = bezier(i * 0.01, points)
        full_list[x] = y
    if index_x in full_list:
        return full_list[index_x]

    closest_number = take_closest(index_x, list(full_list.keys()))
    return full_list[closest_number]


def bezier_curve_range(n, points):
    for i in range(n):
        t = i / float(n - 1)
        yield bezier(t, points)


def get_data_on_percentage(percentage, points_list):
    base_size = points_list[-1][0]

    return bezier_curve_y_from_x(percentage * base_size, points_list) / base_size



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utility functions related with email
"""

from __future__ import print_function, division, absolute_import

import os
import webbrowser
try:
    import urllib2 as urllib
except ImportError:
    import urllib

from tpDcc.libs.python import osplatform


def open_web(url):
    """
    Open given web URL in user web browser
    :param url: str
    """

    if osplatform.is_linux():
        try:
            os.system('gio open {}'.format(url))
        except Exception:
            webbrowser.open(url, 0)
    else:
        webbrowser.open(url, 0)


def safe_open_url(url):
    """
    Opens given URL in a safe way
    :param url: str
    :return:
    """

    try:
        result = urllib.urlopen(url)
    except urllib.HTTPError as exc:
        raise Exception('{} : {}'.format(exc, exc.url))

    return result



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains generic color functionality
"""

from __future__ import print_function, division, absolute_import

import math
import colorsys

from tpDcc.libs.python import mathlib


def convert_hsv_to_rgb(hsv):
    """
    Converts HSV (0-360, 0-1 ranges) values to RGB (0-1 range in SRGBFloats)
    :param hsv: list, hue, saturation, value in 0-360 range, saturation, value in 0-1 range
    :return: list, red, green, blue RGB values in 0-1 srgbFloat
    """

    return list(colorsys.hsv_to_rgb((hsv[0] / 360.0), hsv[1], hsv[2]))


def convert_rgb_to_hsv(rgb):
    """
    Converts RGB (0-1 range in SRGBFloats) to HSV (0-360, 0-1 ranges) values
    :param rgb: list, red, green, blue RGB values in 0-1 srgbFloat
    :return: list, hue, saturation, value in 0-360 range, saturation, value in 0-1 range
    """

    hsv = list(colorsys.rgb_to_hsv((rgb[0]), rgb[1], rgb[2]))
    hsv[0] *= 360.0

    return hsv


def convert_single_srgb_to_linear(color_value):
    """
    Changes a single RGB color to linear space
    :param color_value: float, a single value in 0-1 range (for example red channel)
    :return: float, new color converted to linear
    """

    a = 0.055
    if color_value <= 0.04045:
        return color_value * (1.0 / 12.92)

    return pow((color_value + a) * (1.0 / (1 + a)), 2.4)


def convert_single_linear_to_srgb(color_value):
    """
    Changes as single RGB color in linear to SRGB color space
    :param color_value:float, single color value in 0-1 range (for example red channel)
    :return:float, new color converted to SRGB
    """

    a = 0.055
    if color_value <= 0.0031308:
        return color_value * 12.92

    return (1 + a) * pow(color_value, 1 / 2.4) - a


def convert_color_srgb_to_linear(srgb_color):
    """
    Changes a SRGB color to linear color
    :param srgb_color: list(float, float, float), SRGB float color in list/tuple in 0-1 range
    :return: tuple(float, float, float), new color gamma converted to linear
    """

    return (
        convert_single_srgb_to_linear(srgb_color[0]),
        convert_single_srgb_to_linear(srgb_color[1]),
        convert_single_srgb_to_linear(srgb_color[2])
    )


def convert_color_linear_to_srgb(linear_rgb):
    """
    Changes a linear color to SRGB color
    :param linear_rgb: list(float, float, float), RGB color list/tuple in 0-1 range
    :return: tuple(float, float, flaot), new color gamma converted to SRGB
    """

    return (
        convert_single_linear_to_srgb(linear_rgb[0]),
        convert_single_linear_to_srgb(linear_rgb[1]),
        convert_single_linear_to_srgb(linear_rgb[2])
    )


def hex_to_rgba(hex_str):
    """
    Converts hexadecimal number to RGBA tuple in following formats:
        - "RRGGBB" (2F2F2F)
        - "AARRGGBB" (882F2F2F)
        - "RGB" (CCC)
    :param hex_str: str, hexadecimal string
    :return: tuple(float, float, float, float), color tuple (R, G, B, A)
    """

    if hex_str.startswith('#'):
        hex_str = hex_str[1:]

    if len(hex_str) == 8:
        return int(hex_str[2:4], 16), int(hex_str[4:6], 16), int(hex_str[6:8], 16), int(hex_str[0:2], 16)
    elif len(hex_str) == 6:
        return int(hex_str[0:2], 16), int(hex_str[2:4], 16), int(hex_str[4:6], 16), 255
    elif len(hex_str) == 3:
        return int(hex_str[0:1] * 2, 16), int(hex_str[1:2] * 2, 16), int(hex_str[2:3] * 2, 16), 255
    else:
        raise Exception('Invalid hex length: {} ({})'.format(hex_str, len(hex_str)))


def hex_to_rgb(hex_str):
    """
    Converts hexadecimal number to RGBA tuple
    :param hex_str: str, hexadecimal string
    :return: tuple(float, float, float), color tuple (R, G, B)
    """

    return hex_to_rgba(hex_str)[0:3]


def rgb_to_hex(rgb):
    """
    Converts RGB tuple to hexadecimal string
    :param rgb: tuple(float, float, float): color tuple (R, G, B)
    :return: str, hexadecimal string (eg. 2F2F2F)
    """

    ret = ''.join([hex(h)[2:].upper().zfill(2) for h in rgb])
    if len(ret) == 8:
        # Move last two characters to the beginning
        return ret[-2] + ret[:2]

    return ret


def rgb_int_to_float(color):
    """
    Turns an integer color in 0-255 range into a 0-1 float range
    :param color: tuple(int, int, int, int), color in 0-255 range
    :return: tuple(float, float, float, float) color in 0-1 range
    """

    return tuple([color_channel / 255.0 for color_channel in color])


def rgb_float_to_int(color):
    """
    Turns a float color in 0-1 range into a 0-255 integer range
    :param color: tuple(float, float, float, float), color in 0-1 range
    :return: tuple(int, int, int, int), color in 0-255 range
    """

    return tuple([int(round(255 * float(color_channel))) for color_channel in color])


def rgb_int_round(color):
    """
    Rounds all values of 255 color (eg. (243.1, 100, 10.3) is returned as (243, 100, 10)
    :param color: tuple(int, int, int, int), int color tuple
    :return: tuple(int, int, int, int), color converted
    """

    return tuple([int(round(color_channel)) for color_channel in color])


def convert_srgb_list_to_linear(srgb_list, round_number=True):
    """
    Converts a list of SRGB colors to linear. Can optionally round the result color to 4 decimals
    :param srgb_list: list, list(list(int, int, int), list of SRGB colors in 0-1 range
    :param round_number: bool, Whether to round result color values to 4 decimals or not
    :return: list(list(int, int, int), list of converted colors
    """

    linear_srgb_list = list()
    for srgb_color in srgb_list:
        linear_color_long = convert_color_srgb_to_linear(srgb_color)
        if round_number:
            linear_color = list()
            for long_number in linear_color_long:
                rounded_number = round(long_number, 4)
                linear_color.append(rounded_number)
        else:
            linear_srgb_list.append(linear_color_long)

    return linear_srgb_list


def hsl_color_offset_float(rgb_color, hue_offset=0, saturation_offset=0, lightness_offset=0):
    """
    Offsets color with hue, saturation and lightness (brighten/darken) values
    :param rgb_color: tuple(float, float, float), RGB color in 0.0-1.0 float range
    :param hue_offset: float, hue offset in 0-360 range
    :param saturation_offset: float, saturation offset in 0-255 range
    :param lightness_offset: float, lightness value offset, lighten(0.2) or darken (-0.3) in -1.0 and 1.0 range
    :return: tuple(float, float, float), color in 0-1 range
    """

    if hue_offset:
        hsv = convert_rgb_to_hsv(list(rgb_color))
        new_hue = hsv[0] + hue_offset
        if new_hue > 360.0:
            new_hue -= 360.0
        elif new_hue < 360.0:
            new_hue += 360.0
        hsv = (new_hue, hsv[1], hsv[2])
        rgb_color = convert_hsv_to_rgb(list(hsv))
    if saturation_offset:
        hsv = convert_rgb_to_hsv(rgb_color)
        hsv = (hsv[0], mathlib.clamp(hsv[1] + saturation_offset), hsv[2])
        rgb_color = convert_hsv_to_rgb(list(hsv))
    if lightness_offset:
        rgb_color = (
            mathlib.clamp(rgb_color[0] + lightness_offset),
            mathlib.clamp(rgb_color[1] + lightness_offset),
            mathlib.clamp(rgb_color[2] + lightness_offset)
        )

    return rgb_color


def hsl_color_offset_int(rgb_color, hue_offset=0, saturation_offset=0, lightness_offset=0):
    """
    Offsets color with hue, saturation and lightness (brighten/darken) values
    :param rgb_color: tuple(float, float, float), RGB color in 0-255 integer range
    :param hue_offset: float, hue offset in 0-360 range
    :param saturation_offset: float, saturation offset in 0-255 range
    :param lightness_offset: float, lightness value offset, lighten(0.2) or darken (-0.3) in -1.0 and 1.0 range
    :return: tuple(float, float, float), color in 0-1 range
    """

    rgb_color = rgb_int_to_float(rgb_color)
    lightness_offset = float(lightness_offset) / 255.0
    saturation_offset = float(saturation_offset) / 255.0
    rgb_color = hsl_color_offset_float(
        rgb_color, hue_offset=hue_offset, saturation_offset=saturation_offset, lightness_offset=lightness_offset)

    return rgb_float_to_int(rgb_color)


def desaturate(color, level=1.0):
    """
    Returns a desaturated color
    :param color: tuple(int, int, int, int), color tuple
    :param level: float, level of desaturation from 0 to 1.0. 1.0 is full desaturation and 0 same saturation
    :return: tuple(int, int, int, int), desaturated color
    """

    saturation = convert_rgb_to_hsv(rgb_int_to_float(color))[1]
    saturation_offset = 0.0
    if saturation:
        saturation_offset = int((saturation * level) * -255.0)
    desaturated = hsl_color_offset_int(color, lightness_offset=-40, saturation_offset=saturation_offset)

    return desaturated


def offset_hue_color(hsv, offset):
    """
    Offsets the hue value in -360-360 range by the given offset amount and keeps range by looping
    :param hsv: list(float, float, float), list or tuple representing the hue saturation and value color
    :param offset: float, offset value to offset the saturation
    :return: list(float, float, float), offset hue saturation value color
    """

    if offset > 360:
        offset = 360
    elif offset < -360:
        offset = -360
    hsv[0] += offset
    if hsv[0] > 360:
        hsv[0] -= 360
    elif hsv[0] < 0:
        hsv[0] += 360

    return hsv


def offset_saturation(hsv, offset):
    """
    Offsets the saturation value in 0-1 range by the given offset amount
    :param hsv: list(float, float, float), list or tuple representing the hue saturation and value color
    :param offset: float, offset value to offset the saturation
    :return: list(float, float, float), offset hue saturation value color
    """

    hsv[1] += offset
    if hsv[1] > 1:
        hsv[1] = 1
    elif hsv[1] < 0:
        hsv[1] = 0

    return hsv


def offset_value(hsv, offset):
    """
    Offsets the value (brightness/darkness) in range 0-1 by the given offset amount
    :param hsv: list(float, float, float), list or tuple representing the hue saturation and value color
    :param offset: float, offset value to offset the color
    :return: list(float, float, float), offset hue saturation value color
    """

    hsv[2] += offset
    if hsv[2] > 1:
        hsv[2] = 1
    elif hsv[2] < 0:
        hsv[2] = 0

    return hsv


def offset_color(color_to_offset, offset=0):
    """
    Returns a color with the offset applied
    :param color_to_offset: tuple(int, int, int), color in for of tuple of 3 digits
    :param offset: int, color offset value
    :return: tuple(int, int, int), offset color
    """

    return tuple([mathlib.clamp((color_channel + offset), 0, 255) for color_channel in color_to_offset])


def hue_shift(color_to_shift, shift_amount):
    """
    Shifts the hue of the given color
    :param color_to_shift: tuple(int, int, int), color to shift
    :param shift_amount: int, distance and direction of the color shift
    :return: tuple(int, int, int), color with shifted hue
    """

    rgb_rotator = RGBRotate()
    rgb_rotator.set_hue_rotation(shift_amount)

    return rgb_rotator.apply(*color_to_shift)


def string_is_hex(color_str):
    """
    Returns whether or not given string is a valid hexadecimal string
    :param color_str: str
    :return: bool
    """

    if color_str.startswith('#'):
        return len(color_str) in (4, 7, 9)
    else:
        return len(color_str) in (3, 6, 8)


def convert_kelvin_to_rgb(color_temperature):
    """
    Converts from Kelvin to RGB
    http://www.tannerhelland.com/4435/convert-temperature-rgb-algorithm-code/
    :param color_temperature: float, color temperate in kelvin degrees
    :return: tuple(int, int, int), SRGB color in 0-255 format
    """

    # range check
    if color_temperature < 1000:
        color_temperature = 1000
    elif color_temperature > 40000:
        color_temperature = 40000

    tmp_internal = color_temperature / 100.0

    # red
    if tmp_internal <= 66:
        red = 255
    else:
        tmp_red = 329.698727446 * math.pow(tmp_internal - 60, -0.1332047592)
        if tmp_red < 0:
            red = 0
        elif tmp_red > 255:
            red = 255
        else:
            red = tmp_red

    # green
    if tmp_internal <= 66:
        tmp_green = 99.4708025861 * math.log(tmp_internal) - 161.1195681661
        if tmp_green < 0:
            green = 0
        elif tmp_green > 255:
            green = 255
        else:
            green = tmp_green
    else:
        tmp_green = 288.1221695283 * math.pow(tmp_internal - 60, -0.0755148492)
        if tmp_green < 0:
            green = 0
        elif tmp_green > 255:
            green = 255
        else:
            green = tmp_green

    # blue
    if tmp_internal >= 66:
        blue = 255
    elif tmp_internal <= 19:
        blue = 0
    else:
        tmp_blue = 138.5177312231 * math.log(tmp_internal - 10) - 305.0447927307
        if tmp_blue < 0:
            blue = 0
        elif tmp_blue > 255:
            blue = 255
        else:
            blue = tmp_blue

    return red, green, blue


class RGBRotate(object):
    """
    Hue Rotation, using the matrix rotation method.
    https://stackoverflow.com/questions/8507885/shift-hue-of-an-rgb-color
    """

    def __init__(self):
        self.matrix = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]

    def set_hue_rotation(self, degrees):
        cos_a = math.cos(math.radians(degrees))
        sin_a = math.sin(math.radians(degrees))
        self.matrix[0][0] = cos_a + (1.0 - cos_a) / 3.0
        self.matrix[0][1] = 1. / 3. * (1.0 - cos_a) - math.sqrt(1. / 3.) * sin_a
        self.matrix[0][2] = 1. / 3. * (1.0 - cos_a) + math.sqrt(1. / 3.) * sin_a
        self.matrix[1][0] = 1. / 3. * (1.0 - cos_a) + math.sqrt(1. / 3.) * sin_a
        self.matrix[1][1] = cos_a + 1. / 3. * (1.0 - cos_a)
        self.matrix[1][2] = 1. / 3. * (1.0 - cos_a) - math.sqrt(1. / 3.) * sin_a
        self.matrix[2][0] = 1. / 3. * (1.0 - cos_a) - math.sqrt(1. / 3.) * sin_a
        self.matrix[2][1] = 1. / 3. * (1.0 - cos_a) + math.sqrt(1. / 3.) * sin_a
        self.matrix[2][2] = cos_a + 1. / 3. * (1.0 - cos_a)

    def apply(self, r, g, b):

        from tpDcc.libs.python import mathlib

        rx = r * self.matrix[0][0] + g * self.matrix[0][1] + b * self.matrix[0][2]
        gx = r * self.matrix[1][0] + g * self.matrix[1][1] + b * self.matrix[1][2]
        bx = r * self.matrix[2][0] + g * self.matrix[2][1] + b * self.matrix[2][2]
        return mathlib.clamp(rx, 0, 255), mathlib.clamp(gx, 0, 255), mathlib.clamp(bx, 0, 255)


def compare_rgb_colors_tolerance(first_rgb_color, second_rgb_color, tolerance):
    """
    Compares to RGB colors taking into account the given tolerance (margin for error)
    :param first_rgb_color: tuple(float, float, float), first color to compare
    :param second_rgb_color: tuple(float, float, float), second color to compare
    :param tolerance: float, range in which the colors can vary
    :return: bool, True if two colors matches within the given tolerance; False otherwise
    """

    for i, value in enumerate(first_rgb_color):
        if not abs(value - second_rgb_color[i]) <= tolerance:
            return False

    return True



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Implementation of an kdtree data structure
http://en.wikipedia.org/wiki/Kd-tree
https://github.com/cryptogoth/skc-python/blob/master/skc/kdtree.py
"""

from __future__ import print_function, division, absolute_import


def square_distance(point_a, pointB):
    # squared euclidean distance
    distance = 0
    dimensions = len(point_a)   # assumes both points have the same dimensions
    for dimension in range(dimensions):
        distance += (point_a[dimension] - pointB[dimension]) ** 2
    return distance


class KDTreeNode(object):
    def __init__(self, point, left, right):
        self.point = point
        self.left = left
        self.right = right

    def is_leaf(self):
        return self.left is None and self.right is None


class KDTreeNeighbours(object):
    """
    Internal structure used in nearest-neighbours search.
    """

    def __init__(self, query_point, t):
        self.query_point = query_point
        self.t = t  # neighbours wanted
        self.largest_distance = 0  # squared
        self.current_best = []

    def calculate_largest(self):
        if self.t >= len(self.current_best):
            self.largest_distance = self.current_best[-1][1]
        else:
            self.largest_distance = self.current_best[self.t - 1][1]

    def add(self, point):
        sd = square_distance(point, self.query_point)
        # run through current_best, try to find appropriate place
        for i, e in enumerate(self.current_best):
            if i == self.t:
                return  # enough neighbours, this one is farther, let's forget it
            if e[1] > sd:
                self.current_best.insert(i, [point, sd])
                self.calculate_largest()
                return
        # append it to the end otherwise
        self.current_best.append([point, sd])
        self.calculate_largest()

    def get_best(self):
        return [element[0] for element in self.current_best[:self.t]]


class KDTree(object):
    """
    KDTree implementation.
        Example usage:
            from kdtree import KDTree

            data = <load data> # iterable of points (which are also iterable, same length)
            point = <the point of which neighbours we're looking for>

            tree = KDTree.construct_from_data(data)
            nearest = tree.query(point, t=4) # find nearest 4 points
    """

    def __init__(self, data):
        def build_kdtree(point_list, depth):
            # code based on wikipedia article: http://en.wikipedia.org/wiki/Kd-tree
            if not point_list:
                return None

            # select axis based on depth so that axis cycles through all valid values
            axis = depth % len(point_list[0])  # assumes all points have the same dimension

            # sort point list and choose median as pivot point,
            # TODO: better selection method, linear-time selection, distribution
            point_list.sort(key=lambda point: point[axis])
            median = int(len(point_list) / 2)  # choose median

            # create node and recursively construct subtrees
            node = KDTreeNode(point=point_list[median],
                              left=build_kdtree(point_list[0:median], depth + 1),
                              right=build_kdtree(point_list[median + 1:], depth + 1))
            return node

        self.root_node = build_kdtree(data, depth=0)

    @staticmethod
    def construct_from_data(data):
        tree = KDTree(data)
        return tree

    def query(self, query_point, t=1):
        statistics = {'nodes_visited': 0, 'far_search': 0, 'leafs_reached': 0}

        def nn_search(node, query_point, t, depth, best_neighbours):
            if node is None:
                return

            # statistics['nodes_visited'] += 1

            # if we have reached a leaf, let's add to current best neighbours,
            # (if it's better than the worst one or if there is not enough neighbours)
            if node.is_leaf():
                # statistics['leafs_reached'] += 1
                best_neighbours.add(node.point)
                return

            # this node is no leaf

            # select dimension for comparison (based on current depth)
            axis = depth % len(query_point)

            # figure out which subtree to search
            near_subtree = None  # near subtree
            far_subtree = None  # far subtree (perhaps we'll have to traverse it as well)

            # compare query_point and point of current node in selected dimension
            # and figure out which subtree is farther than the other
            if query_point[axis] < node.point[axis]:
                near_subtree = node.left
                far_subtree = node.right
            else:
                near_subtree = node.right
                far_subtree = node.left

            # recursively search through the tree until a leaf is found
            nn_search(near_subtree, query_point, t, depth + 1, best_neighbours)

            # while unwinding the recursion, check if the current node
            # is closer to query point than the current best,
            # also, until t points have been found, search radius is infinity
            best_neighbours.add(node.point)

            # check whether there could be any points on the other side of the
            # splitting plane that are closer to the query point than the current best
            if (node.point[axis] - query_point[axis]) ** 2 < best_neighbours.largest_distance:
                # statistics['far_search'] += 1
                nn_search(far_subtree, query_point, t, depth + 1, best_neighbours)

            return

        # if there's no tree, there's no neighbors
        if self.root_node is not None:
            neighbours = KDTreeNeighbours(query_point, t)
            nn_search(self.root_node, query_point, t, depth=0, best_neighbours=neighbours)
            result = neighbours.get_best()
        else:
            result = []

        # print statistics
        return result



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains functions and classes related with sqlite module
"""

from __future__ import print_function, division, absolute_import

import sqlite3


class ConnectionContext(object):
    """
    Contextual class that handles the connection and cursor mechanism. Its intended to be an optimisation to allow
    for multiple calls to occur without connection overheads
    """

    def __init__(self, identifier, commit=False, get=False):

        self._identifier = identifier
        self._commit = commit
        self._get = get

        self._connection = None
        self._cursor = None
        self._results = list()

    def __enter__(self):
        self._connection = sqlite3.connect(self._identifier)
        self._cursor = self._connection.cursor()
        self._cursor.execute("PRAGMA foreign_keys = ON")

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._get:
            self._results = self._cursor.fetchall()

        if self._commit:
            self._connection.commit()

        self._connection.close()

    @property
    def cursor(self):
        return self._cursor

    @property
    def results(self):
        return self._results



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utility contexts
"""

from __future__ import print_function, division, absolute_import

import time
import logging
import contextlib

LOGGER = logging.getLogger('tpDcc-libs-python')


@contextlib.contextmanager
def empty_decorator_context():
    """
    Empty context decorator
    """

    pass


class Timer(object):
    def __init__(self, description, logger=None):
        self._start = 0.0
        self._end = 0.0
        self._description = description
        self._logger = logger or LOGGER

    def __enter__(self):
        self._start = time.time()

    def __exit__(self, *args, **kwargs):
        self._end = time.time()
        self._logger.info("{}: {}".format(self._description, (self._end - self._start)))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utility functions related with apps
"""


from __future__ import print_function, division, absolute_import

import sys


def is_nuke():
    """
    Checks if Nuke is available or not
    :return: bool
    """

    try:
        import nuke
        return True
    except ImportError:
        return False


def is_maya():
    """
    Checks if Maya is available or not
    :return: bool
    """

    return 'maya.exe' in sys.executable.lower()


def is_mayapy():
    """
    Checks if Maya is available or not
    :return: bool
    """

    return 'mayapy.exe' in sys.executable.lower()


def is_max():
    """
    Checks if Max is available or not
    :return: bool
    """

    return '3dsmax' in sys.executable.lower()


def is_houdini():
    """
    Checks if Houdini is available or not
    :return: bool
    """

    return 'houdini' in sys.executable


def is_motionbuilder():
    """
    Checks if MotionBuilder is available or not
    :return: bool
    """

    return 'motionbuilder' in sys.executable.lower()



# !/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

"""
Module that contains functions to work with win32
"""

import sys

if 'win' in sys.platform:
    import ctypes
    import ctypes.wintypes

from tpDcc.libs.python import python

GWL_WNDPROC = -4
GWL_HINSTANCE = -6
GWL_HWNDPARENT = -8
GWL_STYLE = -16
GWL_EXSTYLE = -20
GWL_USERDATA = -21
GWL_ID = -12

WS_BORDER = 0x800000
WS_CAPTION = 0xc00000
WS_CHILD = 0x40000000
WS_CHILDWINDOW = 0x40000000
WS_CLIPCHILDREN = 0x2000000
WS_CLIPSIBLINGS = 0x4000000
WS_DISABLED = 0x8000000
WS_DLGFRAME = 0x400000
WS_GROUP = 0x20000
WS_HSCROLL = 0x100000
WS_ICONIC = 0x20000000
WS_MAXIMIZE = 0x1000000
WS_MAXIMIZEBOX = 0x10000
WS_MINIMIZE = 0x20000000
WS_MINIMIZEBOX = 0x20000
WS_OVERLAPPED = 0
WS_OVERLAPPEDWINDOW = 0xcf0000
WS_POPUP = 0x80000000
WS_POPUPWINDOW = 0x80880000
WS_SIZEBOX = 0x40000
WS_SYSMENU = 0x80000
WS_TABSTOP = 0x10000
WS_THICKFRAME = 0x40000
WS_TILED = 0
WS_TILEDWINDOW = 0xcf0000
WS_VISIBLE = 0x10000000
WS_VSCROLL = 0x200000


def to_hwnd(pycobject):
    """
    Convenience method to get a Windows Handle from a PySide WinID
    Based on http://srinikom.github.io/pyside-bz-archive/523.html
    @return A value equivalent to a void* that represents the Windows handle if one exists; None otherwise.
    """

    if type(pycobject) is int:
        # That specific case happen in maya 2017, here, we already have the hwnd so no further manipulation is needed
        return pycobject
    if sys.version_info[0] == 2:
        ctypes.pythonapi.PyCObject_AsVoidPtr.restype = ctypes.c_void_p
        ctypes.pythonapi.PyCObject_AsVoidPtr.argtypes = [ctypes.py_object]
        return ctypes.pythonapi.PyCObject_AsVoidPtr(pycobject)
    elif sys.version_info[0] == 3:
        ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_void_p
        ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object]
        return ctypes.pythonapi.PyCapsule_GetPointer(pycobject, None)


def set_owner(hwnd, hwnd_owner):
    """
    Changes the owner window of the given window
    :param hwnd:
    :param hwnd_owner:
    """

    _update_window = ctypes.windll.user32.UpdateWindow

    # WIN32 vs WIN64 - from a macro in winuser.h
    if ctypes.sizeof(ctypes.wintypes.HWND) == ctypes.sizeof(ctypes.c_long):
        _LONG = ctypes.wintypes.LONG
        _set_window_long = ctypes.windll.user32.SetWindowLongW
        _set_window_long.argtypes = [ctypes.wintypes.HWND, ctypes.c_int, ctypes.wintypes.LONG]
        _set_window_long.restype = ctypes.c_void_p
    elif ctypes.sizeof(ctypes.wintypes.HWND) == ctypes.sizeof(ctypes.c_longlong):
        _LONG = ctypes.wintypes.HWND
        _set_window_long = ctypes.windll.user32.SetWindowLongPtrW
        _set_window_long.argtypes = [ctypes.wintypes.HWND, ctypes.c_int, ctypes.wintypes.HWND]
        _set_window_long.restype = _LONG

    last_error = ctypes.set_last_error(0)
    try:
        result = _set_window_long(ctypes.wintypes.HWND(hwnd), ctypes.c_int(GWL_HWNDPARENT), _LONG(hwnd_owner))
    finally:
        last_error = ctypes.set_last_error(last_error)

    if not result and last_error:
        raise ctypes.WinError(last_error)

    _update_window(hwnd_owner)

    return result


def get_reg_key(registry, key, architecture=None):
    """
    Returns a _winreg hkey if found
    :param registry: str, registry to look in. HKEY_LOCAL_MACHINE for example
    :param key: str, key to open 'Software/Ubisoft/Test' for example
    :param architecture: variant, int || None, 32 or 64 bit. If None, default system architecture is used
    :return: _winreg handle object
    """

    try:
        import _winreg as winreg
    except ImportError:
        import winreg

    reg_key = None
    a_reg = winreg.ConnectRegistry(None, getattr(winreg, registry))
    if architecture == 32:
        sam = winreg.KEY_WOW64_32KEY
    elif architecture == 64:
        sam = winreg.KEY_WOW64_64KEY
    else:
        sam = 0
    try:
        reg_key = winreg.OpenKey(a_reg, key, 0, winreg.KEY_READ | sam)
    except WindowsError:
        pass

    return reg_key


def list_reg_keys(registry, key, architecture=None):
    """
    Returns a list of child keys as tuples containing:
        - A string that identifies the value name
        - An object that holds the value data, and whose type depends on the underlying registry type
        - An integer that identifies the type of the value data (see table in docs for _winreg.SetValueEx)
    :param registry: str, registry to look in. HKEY_LOCAL_MACHINE for example
    :param key: str, key to open 'Software/Ubisoft/Test' for example
    :param architecture: variant, int || None, 32 or 64 bit. If None, default system architecture is used
    :return: list<tuple>
    """

    try:
        import _winreg as winreg
    except ImportError:
        import winreg

    reg_key = get_reg_key(registry=registry, key=key, architecture=architecture)
    ret = list()
    if reg_key:
        i = 0
        while True:
            try:
                ret.append(winreg.EnumKey(reg_key, i))
                i += 1
            except WindowsError:
                break

    return ret


def list_reg_key_values(registry, key, architecture=None):
    """
    Returns a list of child keys and their values as tuples containing:
        - A string that identifies the value name
        - An object that holds the value data, and whose type depends on the underlying registry type
        - An integer that identifies the type of the value data (see table in docs for _winreg.SetValueEx)
    :param registry: str, registry to look in. HKEY_LOCAL_MACHINE for example
    :param key: str, key to open 'Software/Ubisoft/Test' for example
    :param architecture: variant, int || None, 32 or 64 bit. If None, default system architecture is used
    :return: list<tuple>
    """

    try:
        import _winreg as winreg
    except ImportError:
        import winreg

    reg_key = get_reg_key(registry=registry, key=key, architecture=architecture)
    ret = list()
    if reg_key:
        sub_keys, value_count, modified = winreg.QueryInfoKey(reg_key)
        for i in range(value_count):
            ret.append(winreg.EnumValue(reg_key, i))

    return ret


def registry_value(registry, key, value_name, architecture=None):
    """
    Retruns the value and type of the given registry key value name
    :param registry: str, registry to look in. HKEY_LOCAL_MACHINE for example
    :param key: str, key to open 'Software/Ubisoft/Test' for example
    :param value_name: str, name of the value to read. To read the 'default' key, pass an empty string
    :param architecture: variant, int || None, 32 or 64 bit. If None, default system architecture is used
    :return: tuple<object, int, value stored in key and registry type for value (see _winreg's Value Types)
    """

    reg_key = get_reg_key(registry, key, architecture=architecture)
    if reg_key:
        import _winreg
        value = _winreg.QueryValueEx(reg_key, value_name)
        _winreg.CloseKey(reg_key)
        return value

    return '', 0


def get_monitors():
    """
    Returns a list of all monitors
    code.activestate.com/recipes/460509-get-the-actual-and-usable-sizes-of-all-the-monitor
    :return:
    """

    result = list()

    CBFUNC = ctypes.WINFUNCTYPE(ctypes.c_int, ctypes.c_ulong, ctypes.c_ulong, ctypes.POINTER(Rect), ctypes.c_double)

    def cb(h_monitor, hdc_monitor, l_prc_monitor, dw_data):
        r = l_prc_monitor.contents
        data = [h_monitor]
        data.append(r.dump())
        result.append(data)

        return 1

    cb_fn = CBFUNC(cb)
    ctypes.windll.user32.EnumDisplayMonitors(0, 0, cb_fn, 0)

    return result


def get_active_monitor_area():
    """
    Returns the active and working area of each monitor
    code.activestate.com/recipes/460509-get-the-actual-and-usable-sizes-of-all-the-monitor
    :return:
    """

    result = list()
    monitors = get_monitors()
    for h_monitor, extents in monitors:
        data = [h_monitor]
        monitor_info = MonitorInfo()
        monitor_info.cbSize = ctypes.sizeof(MonitorInfo)
        monitor_info.rcMonitor = Rect()
        monitor_info.rcWork = Rect()
        ctypes.windll.user32.GetMonitorInfoA(h_monitor, ctypes.byref(monitor_info))
        data.append(monitor_info.rcMonitor.dump())
        data.append(monitor_info.rcWork.dump())
        result.append(data)

    return result


def focus_window_from_pid(window_pid, restore=True):

    def _window_enumeration_handler(hwnd, list_to_append):
        list_to_append.append((hwnd, win32gui.GetWindowText(hwnd)))

    import win32gui

    hwnds = get_hwnds_for_pid(window_pid)

    window_list = []
    win32gui.EnumWindows(_window_enumeration_handler, window_list)
    for i in window_list:
        if i[0] in hwnds:
            show_type = 9 if restore else 5
            win32gui.ShowWindow(i[0], show_type)
            win32gui.SetForegroundWindow(i[0])
            break


def get_hwnds_for_pid(pid):

    import win32gui
    import win32process

    def _window_enumeration_handler(hwnd, hwnds):
        if win32gui.IsWindowVisible(hwnd) and win32gui.IsWindowEnabled(hwnd):
            _, found_pid = win32process.GetWindowThreadProcessId(hwnd)
            if found_pid == pid:
                hwnds.append(hwnd)
        return True

    hwnds = []
    win32gui.EnumWindows(_window_enumeration_handler, hwnds)
    return hwnds


def _get_win_folder_from_registry(csidl_name):
    """
    Based on appdirs _get_win_folder_from_registry function
    """

    if python.is_python3():
        import winreg as _winreg
    else:
        import _winreg

    shell_folder_name = {
        "CSIDL_APPDATA": "AppData",
        "CSIDL_COMMON_APPDATA": "Common AppData",
        "CSIDL_LOCAL_APPDATA": "Local AppData",
    }[csidl_name]

    key = _winreg.OpenKey(
        _winreg.HKEY_CURRENT_USER,
        r"Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders"
    )
    dir, type = _winreg.QueryValueEx(key, shell_folder_name)

    return dir


def _get_win_folder_with_ctypes(csidl_name):
    """
    Based on appdirs _get_win_folder_with_ctypes function
    """

    csidl_const = {
        "CSIDL_APPDATA": 26,
        "CSIDL_COMMON_APPDATA": 35,
        "CSIDL_LOCAL_APPDATA": 28,
    }[csidl_name]

    buf = ctypes.create_unicode_buffer(1024)
    ctypes.windll.shell32.SHGetFolderPathW(None, csidl_const, None, 0, buf)

    # Downgrade to short path name if have highbit chars. See
    # <http://bugs.activestate.com/show_bug.cgi?id=85099>.
    has_high_char = False
    for c in buf:
        if ord(c) > 255:
            has_high_char = True
            break
    if has_high_char:
        buf2 = ctypes.create_unicode_buffer(1024)
        if ctypes.windll.kernel32.GetShortPathNameW(buf.value, buf2, 1024):
            buf = buf2

    return buf.value


if 'win' in sys.platform:
    try:
        from ctypes import windll
        get_win_folder = _get_win_folder_with_ctypes
    except ImportError:
        get_win_folder = _get_win_folder_from_registry

    class Rect(ctypes.Structure):
        _fields_ = [
            ('left', ctypes.c_long),
            ('top', ctypes.c_long),
            ('right', ctypes.c_long),
            ('bottom', ctypes.c_long)
        ]

        def dump(self):
            return tuple(map(int, (self.left, self.top, self.right, self.bottom)))

    class MonitorInfo(ctypes.Structure):
        _fields_ = [
            ('cbSize', ctypes.c_long),
            ('rcMonitor', Rect),
            ('rcWork', Rect),
            ('dwFlags', ctypes.c_long)
        ]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementations for Composition mechanism
https://github.com/mikemalinowski/xcomposite
"""

from __future__ import print_function, division, absolute_import


class Composition(object):

    def __init__(self):

        self._components = list()

    def __repr__(self):

        if not self._components:
            return self.__class__.__name__

        return '[{} ({})]'.format(
            self.__class__.__name__, ';'.join(
                [component.__class__.__name__ for component in self._components if component != self]))

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def components(self):
        """
        Returns all components
        :return: list(instance)
        """

        return self._components

    def bind(self, component):
        """
        Adds a component to the class. At this point, all decorated class will incorporate this component
        :param component: class, component to add
        """

        self._components.append(component)

    def unbind(self, component_class):
        """
        Removes the component of given type from the list of components
        :param component_class: class, component class type to remove
        """

        for component in self._components[:]:
            if isinstance(component, component_class):
                self._components.remove(component)


class Ignore(object):
    """
    Class that can be used as a return item when you do not want the value of a function to contribute a bound
    result. Useful when wanting to apply the composite pattern to classes without altering hierarchy or applying
    decorators directly
    """

    pass


def _methods(composition_class, method_name):
    """
    Function for getting a list of all the methods which requires calling.
    """

    return [getattr(component, method_name) for component in composition_class.components()]


def _results(composition_class, method_name, *args, **kwargs):
    """
    Convenience function for return all the results for the methods with the given name on the class.

    :param composition_class: Composition
    :param method_name: Name of method to call
    :param args: Args to pass to call
    :param kwargs: Keyword arguments to pass

    :return: List of results
    """

    results = list()

    for method in _methods(composition_class, method_name):
        result = method(*args[1:], **kwargs)

        if not isinstance(result, Ignore):
            results.append(result)

    return results

# ============================================================================================================
# DECORATORS (List of decorators that can be used in composition functions)
# ============================================================================================================


def take_min(func):
    """
    This decorator assumes a numeric return from each method and will
    return the smallest value.
    """

    def inner(*args, **kwargs):
        return min(
            _results(
                args[0],
                func.__name__,
                *args,
                **kwargs
            )
        )

    return inner


def take_max(func):
    """
    This decorator assumes a numeric return from each method and will
    return the highest value.
    """

    def inner(*args, **kwargs):
        return max(
            _results(
                args[0],
                func.__name__,
                *args,
                **kwargs
            )
        )

    return inner


def take_sum(func):
    """
    This decorator assumes a numeric return from each method and will
    return the sum of all the values.
    """

    def inner(*args, **kwargs):
        return sum(
            _results(
                args[0],
                func.__name__,
                *args,
                **kwargs
            )
        )

    return inner


def take_average(func):
    """
    This decorator assumes a numeric return from each method and will
    return the average (mean) of all the values.
    """

    def inner(*args, **kwargs):
        results = _results(
            args[0],
            func.__name__,
            *args,
            **kwargs
        )

        sum_of_results = sum(results)

        if sum_of_results == 0:
            return sum_of_results

        return sum_of_results / len(results)

    return inner


def take_first(func):
    """
    This decorator will return the first item returned from any of the composited methods.
    """
    def inner(*args, **kwargs):
        for method in _methods(args[0], func.__name__):
            result = method(*args[1:], **kwargs)

            if not isinstance(result, Ignore):
                return result

        return None

    return inner


def first_true(func):
    """
    This decorator will return the first item returned from any of the composited methods.
    """
    def inner(*args, **kwargs):
        for method in _methods(args[0], func.__name__):
            result = method(*args[1:], **kwargs)

            if result and not isinstance(result, Ignore):
                return result

        return None

    return inner


def take_last(func):
    """
    This decorator will return the last item returned from any of the composited methods.
    """

    def inner(*args, **kwargs):
        for method in reversed(_methods(args[0], func.__name__)):
            result = method(*args[1:], **kwargs)

            if not isinstance(result, Ignore):
                return result

        return None

    return inner


def extend_results(func):
    """
    This decorator assumes all returns are lists and will use list.extend on each list given resulting in a single
    list of all results.
    """

    def inner(*args, **kwargs):
        extended_results = list()
        results = _results(
            args[0],
            func.__name__,
            *args,
            **kwargs
        )

        for result in results:
            if not isinstance(result, Ignore):
                extended_results.extend(result)

        return extended_results

    return inner


def extend_unique(func):
    """
    This decorator assumes all returns are lists and will use list.extend on each list given resulting in a single
    list of all results.
    """

    def inner(*args, **kwargs):
        extended_results = list()
        results = _results(
            args[0],
            func.__name__,
            *args,
            **kwargs
        )

        for result in results:
            if not isinstance(result, Ignore):
                extended_results.extend(result)

        return list(set(extended_results))

    return inner


def update_dictionary(func):
    """
    This decorator will update each dictionary results in order
    """

    def inner(*args, **kwargs):
        output = dict()
        results = _results(
            args[0],
            func.__name__,
            *args,
            **kwargs
        )

        for result in results:
            if not isinstance(result, Ignore):
                output.update(result)

        return output

    return inner


def update_dictionary_unique(func):
    """
    This decorator will update each dictionary results in order but will keep keys that were already added in dict
    :param func:
    :return:
    """

    def inner(*args, **kwargs):
        output = dict()
        results = _results(
            args[0],
            func.__name__,
            *args,
            **kwargs
        )

        for result in results:
            if not isinstance(result, Ignore):
                for key, value in result.items():
                    if key in output:
                        continue
                    output[key] = value

        return output

    return inner


def absolute_false(func):
    """
    Returns False if all elements evaluate to False, otherwise True
    is returned.
    """

    def inner(*args, **kwargs):
        results = _results(
            args[0],
            func.__name__,
            *args,
            **kwargs
        )

        for result in results:
            if result:
                return True
        return False

    return inner


def absolute_true(func):
    """
    Returns True if all elements evaluate to True, otherwise False
    is returned.
    """

    def inner(*args, **kwargs):
        results = _results(
            args[0],
            func.__name__,
            *args,
            **kwargs
        )

        for result in results:
            if not result:
                return False
        return True

    return inner


def any_false(func):
    """
    If any items are False, then false is returned.
    """

    def inner(*args, **kwargs):
        results = _results(
            args[0],
            func.__name__,
            *args,
            **kwargs
        )

        for result in results:
            if not result:
                return False
        return True

    return inner


def any_true(func):
    """
    If any items are True, then True is returned.
    """

    def inner(*args, **kwargs):
        results = _results(
            args[0],
            func.__name__,
            *args,
            **kwargs
        )

        for result in results:
            if result:
                return True
        return False

    return inner


def append_results(func):
    """
    This decorator will append each result - regardless of type - into a
    list.
    """

    def inner(*args, **kwargs):
        return _results(
            args[0],
            func.__name__,
            *args,
            **kwargs
        )

    return inner


def append_unique(func):
    """
    This decorator will append each result - regardless of type - into a
    list.
    """

    def inner(*args, **kwargs):
        return list(
            set(
                _results(
                    args[0],
                    func.__name__,
                    *args,
                    **kwargs
                )
            )
        )

    return inner


def take_range(func):
    """
    Returns the range of all the values (max - min). If only one value is given the range will be zero.
    """

    def inner(*args, **kwargs):
        results = _results(
            args[0],
            func.__name__,
            *args,
            **kwargs
        )

        return float(max(results)) - float(min(results))

    return inner



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility methods related to string paths
"""


from __future__ import print_function, division, absolute_import

import os
import stat
import string
import shutil
import logging
import tempfile
import traceback
import contextlib

from tpDcc.libs.python import name, folder, osplatform, python, win32

SEPARATOR = '/'
BAD_SEPARATOR = '\\'
PATH_SEPARATOR = '//'
SERVER_PREFIX = '\\'
RELATIVE_PATH_PREFIX = './'
BAD_RELATIVE_PATH_PREFIX = '../'
WEB_PREFIX = 'https://'

# We use one separator depending if we are working on Windows (nt) or other operative system
NATIVE_SEPARATOR = (SEPARATOR, BAD_SEPARATOR)[os.name == 'nt']

LOGGER = logging.getLogger('tpDcc-libs-python')


class FindUniquePath(name.FindUniqueString, object):
    def __init__(self, directory):
        if not directory:
            directory = folder.get_current_working_directory()

        self.parent_path = self._get_parent_path(directory)
        basename = get_basename(directory=directory)

        super(FindUniquePath, self).__init__(basename)

    def _get_scope_list(self):
        return folder.get_files_and_folders(directory=self.parent_path)

    def _search(self):
        name = super(FindUniquePath, self)._search()
        return join_path(self.parent_path, name)

    def _get_parent_path(self, directory):
        return get_dirname(directory)


@contextlib.contextmanager
def cd(new_dir, cleanup=lambda: True):
    prev_dir = os.getcwd()
    os.chdir(os.path.expanduser(new_dir))
    try:
        yield
    finally:
        os.chdir(prev_dir)
        cleanup()


@contextlib.contextmanager
def temp_dir():
    dir_path = tempfile.mkdtemp()

    def cleanup():
        shutil.rmtree(dir_path)
    with cd(dir_path, cleanup):
        yield dir_path


def normalize_path(path):
    """
    Normalizes a path to make sure that path only contains forward slashes
    :param path: str, path to normalize
    :return: str, normalized path
    """

    path = path.replace(BAD_SEPARATOR, SEPARATOR).replace(PATH_SEPARATOR, SEPARATOR)

    if python.is_python2():
        try:
            path = unicode(path.replace(r'\\', r'\\\\'), "unicode_escape").encode('utf-8')
        except TypeError:
            path = path.replace(r'\\', r'\\\\').encode('utf-8')

    return path.rstrip('/')


def normalize_paths(paths):
    """
    Normalize all the given paths into a consistent format
    :param paths: list(str)
    :return: list(str)
    """

    return [normalize_path(path) for path in paths]


def clean_path(path):
    """
    Cleans a path. Useful to resolve problems with slashes
    :param path: str
    :return: str, clean path
    """

    if not path:
        return

    # We convert '~' Unix character to user's home directory
    path = os.path.expanduser(str(path))

    # Remove spaces from path and fixed bad slashes
    path = normalize_path(path.strip())

    # Fix server paths
    is_server_path = path.startswith(SERVER_PREFIX)
    while SERVER_PREFIX in path:
        path = path.replace(SERVER_PREFIX, PATH_SEPARATOR)
    if is_server_path:
        path = PATH_SEPARATOR + path

    # Fix web paths
    if not path.find(WEB_PREFIX) > -1:
        path = path.replace(PATH_SEPARATOR, SEPARATOR)

    return path


def real_path(path):
    """
    Returns the given path removing any symbolic link
    :param path: str
    :return: str
    """

    path = os.path.realpath(path)
    path = os.path.expanduser(path)

    return normalize_path(path)


def join_path(dir1, dir2):
    """
    Appends dir2 to the end of dir1
    :param dir1: str
    :param dir2: str
    :return: str, combined directory path
    """

    if dir1 is None or dir2 is None:
        return None

    dir1 = clean_path(dir1)
    dir2 = clean_path(dir2)

    dir_path = '{}/{}'.format(dir1, dir2) if dir2 else dir1
    dir_path = clean_path(dir_path)

    return dir_path


def set_windows_slashes(directory):
    """
    Set all the slashes in a name so they use Windows slashes (\)
    :param directory: str
    :return: str
    """

    return directory.replace('/', '\\').replace('//', '\\')


def split_path(path):
    """
    Split the given path into directory, basename and extension
    :param path:
    :return: list(str)
    """

    path = normalize_path(path)
    filename, extension = os.path.splitext(path)

    return os.path.dirname(filename), os.path.basename(filename), extension


def get_relative_path(path, start):
    """
    Gets a relative path from a start path
    :param path: str, path to get relative path
    :param start: str, Start path to calculate the relative path from
    """

    # if os.path.splitext(start)[-1]:
    #     start = clean_path(os.path.dirname(start))
    # rel_path = clean_path(os.path.relpath(path, start))
    #
    # # TODO: Check if this is correct
    # if not rel_path.startswith('../'):
    #     rel_path = './' + rel_path
    #
    # return rel_path

    rpath = start

    for i in range(0, 3):

        rpath = os.path.dirname(rpath)
        token = os.path.relpath(rpath, start)

        rpath = normalize_path(rpath)
        token = normalize_path(token)

        if rpath.endswith("/"):
            rpath = rpath[:-1]

        path = path.replace(rpath, token)

    return path


def get_absolute_path(path, start):
    """
    Gets an absolute path from a start path
    :param path: str, path to get absolute path
    :param start: str, Start path to calculate the absolute path from
    """

    path = path.replace('\\', '/')
    if not os.path.isdir(start):
        start = os.path.dirname(start).replace('\\', '/')
    else:
        start = start.replace('\\', '/')

    return os.path.abspath(os.path.join(start, path)).replace('\\', '/')


def get_absolute_file_paths(root_directory):
    """
    Returns a generator with all absolute paths on a folder (and sub folders)
    :param root_directory: str, directory to start looking
    """

    for root, _, files in os.walk(root_directory):
        for f in files:
            yield os.path.abspath(os.path.join(root, f))


def get_immediate_subdirectories(root_directory):
    """
    Returns a list with intermediate subdirectories of root directory
    :param root_directory: str, directory to start looking
    """

    return [
        os.path.join(
            root_directory, name) for name in os.listdir(root_directory) if os.path.isdir(
            os.path.join(root_directory, name))
    ]


def get_extension(path):
    """
    Returns the exctension of a file path (wihtout the period)
    :param path: str, valid path to a file
    :return: str
    """

    return os.path.splitext(path)[1][1:]


def exists(directory):
    """
    Returns true if the given path exists
    :param directory: str
    :return: bool
    """

    if not directory:
        return False

    try:
        stat = os.stat(directory)
        if stat:
            return True
    except Exception:
        return False

    return os.path.exists(directory)


def has_extension(path, file_extension):
    """
    Checks if a given file path has a specific given file extension
    :param path: str, file path
    :param file_extension: str, valid file extension
    :return: bool, True if the extension of the given path matches the given extension, False otherwise
    """

    return True if get_extension(path=path) == file_extension else False


def get_files(root, file_extension=None, recursive=False, full_path=False, stdout=False):
    """
    Get all files from a given directory
    :param root: str, path to get directories from
    :param file_extension: str, file extension of files to search for
    :param recursive: bool, True if the function will search deeper than one level of files
    :param full_path: str, the output of the path will be the full path if True
    :param stdout: print results in Python output if True
    :return: list<str>, list of files
    """

    def out(data):
        for i in data:
            print(i)
            print('Found {0} files'.format(len(data)))

    if len(root):
        root = clean_path(root)
        directories = [d for d in os.listdir(root) if is_dir(root, d)]
        if file_extension:
            if full_path:
                files = [clean_path(os.path.abspath(os.path.join(root, f))) for
                         f in os.listdir(root) if is_file(root, f) and has_extension(f, file_extension)]
            else:
                files = [f for f in os.listdir(root) if is_file(root, f) and has_extension(f, file_extension)]
        else:
            if full_path:
                files = [
                    clean_path(os.path.abspath(os.path.join(root, f))) for f in os.listdir(root) if is_file(root, f)]
            else:
                files = [f for f in os.listdir(root) if is_file(root, f)]

        if len(directories) and recursive:
            more_files = [get_files(os.path.join(root, d), file_extension, recursive, full_path) for d in directories]
            if len(more_files):
                for chunk in more_files:
                    files.extend(chunk)

        output = [clean_path(p) for p in files]
        if stdout:
            out(output)

        return output
    return None


def get_folders_from_path(path):
    """
    Gets a list of sub folders in the given path
    :param path: str
    :return: list<str>
    """

    folders = list()
    while True:
        path, folder = os.path.split(path)
        if folder != '':
            folders.append(folder)
        else:
            if path != '':
                folders.append(path)
            break
    folders.reverse()

    return folders


def get_basename(directory, with_extension=True):
    """
    Get the last part of a directory name
    For example, C:/test/rig.py will return rig.py if with_extension is True of rig if with_extension is False
    :param directory: str
    :param with_extension: bool, Whether to return the file name with extension or not
    :return: variant, str || bool (if fails)
    """

    from tpDcc.libs.python import fileio

    try:
        base_name = os.path.basename(directory)
        if not with_extension:
            base_name = fileio.remove_extension(base_name)

        return base_name
    except Exception:
        return False


def get_dirname(directory):
    """
    Given a directory path, this will return the path above the last child path in the path
    For example, C:/test/rig.py will return C:/test
    :param directory: str,
    :return: variant, str || bool (if fails)
    """

    try:
        return os.path.dirname(directory)
    except Exception:
        return False


def unique_path_name(directory, padding=0):
    """
    Add padding to the given path name if it is not unique
    :param directory: str, diretory name including path
    :param padding: int, where the padding should start
    :return: str, new unique directory with path
    """

    unique_path = FindUniquePath(directory)
    unique_path.set_padding(padding)

    return unique_path.get()


def get_common_path(path1, path2):
    """
    Returns path that is common in both given paths
    :param path1: str
    :param path2: str
    :return: str, common path shared by both given paths
    """

    path1 = clean_path(path1)
    path2 = clean_path(path2)

    split_path1 = path1.split('/')
    split_path2 = path2.split('/')

    first_list = split_path1
    second_list = split_path2

    found = list()
    for i in range(len(first_list)):
        if len(second_list) <= i:
            break
        if first_list[i] == second_list[i]:
            found.append(first_list[i])
        if first_list[i] != second_list[i]:
            break

    found = string.join(found, '/')

    return found


def remove_common_path(path1, path2):
    """
    Removes path that is common in both given paths
    :param path1: str
    :param path2: str
    :return: str, path without the path shared by both given paths
    """

    path1 = clean_path(path1)
    path2 = clean_path(path2)

    split_path1 = path1.split('/')
    split_path2 = path2.split('/')

    skip = True
    new_path = list()

    for i in range(len(split_path2)):
        if skip:
            if len(split_path1) > i:
                if split_path1[i] != split_path2[i]:
                    skip = False
            if (len(split_path1) - 1) < i:
                skip = False
        if not skip:
            new_path.append(split_path2[i])

    new_path = '/'.join(new_path)

    return new_path


def remove_common_path_at_beginning(path1, path2):
    """
    Removes path that is similar on both given paths at the beginning of both of them
    :param path1: str
    :param path2: str
    :return: str
    """

    path2 = path2 or ''

    value = path2.find(path1)
    sub_part = None

    if value > -1 and value == 0:
        sub_part = path2[len(path1):]
    if sub_part:
        if sub_part.startswith('/'):
            sub_part = sub_part[1:]

    return sub_part


def is_dir(directory, path=None):
    """
    Checks if the given directory is a directory or not
    :param directory: str
    :param path: str
    :return: bool
    """

    if not directory:
        return False

    if path is not None:
        directory = join_path(directory, path)

    try:
        mode = os.stat(directory)[stat.ST_MODE]
        if stat.S_ISDIR(mode):
            return True
    except Exception:
        return False

    return False


def is_file(file_path, path=None):
    """
    Checks if the given path is an existing file
    :param file_path: str
    :return: bool
    """

    if not file_path:
        return False

    if path is not None:
        file_path = join_path(file_path, path)

    try:
        mode = os.stat(file_path)[stat.ST_MODE]
        if stat.S_ISREG(mode):
            return True
    except Exception:
        return False

    return False


def move(path1, path2):
    """
    Move the folder or file pointed by path1 under the directory path2
    :param path1: str, file or folder including path
    :param path2: str, path where path1 should move to
    :return: bool, Whether the move operation was successful
    """

    try:
        shutil.move(path1, path2)
    except Exception:
        LOGGER.warning('Failed to move {0} to {1}'.format(path1, path2))
        return False

    return True


def rename(directory, name, make_unique=False):
    """
    Renames given with a new name
    :param directory: str, full path to the diretory we want to rename
    :param name: str, new name of the folder we want to rename
    :param make_unique: bool, Whether to add a number to the folder name to make it unique
    :return: str, path of the renamed folder
    """

    base_name = get_basename(directory=directory)
    if base_name == name:
        return

    parent_path = get_dirname(directory=directory)
    rename_path = join_path(parent_path, name)

    if make_unique:
        rename_path = unique_path_name(directory=rename_path)

    if is_dir(rename_path) or is_file(rename_path):
        return False

    try:
        os.chmod(directory, 0o777)
        message = 'rename: {0} >> {1}'.format(directory, rename_path)
        LOGGER.info(message)
        os.rename(directory, rename_path)
    except Exception:
        LOGGER.error('{}'.format(traceback.format_exc()))
        return False

    return rename_path


def get_user_data_dir(appname=None, appauthor=None, version=None, roaming=False):
    r"""
    Based on appdirs user_data_dir function

    Returns the full path to the user-specific data directory
    :param appname: str or None, name of the application. If None, the system directory is returned
    :param appauthor: str, name of the author or distributing body for this application. Typically is the owning
        company name. Only used in Windows.
    :param version: str, optional version path element to append to the path. You might want to use this if you
        want multiple versions of you app to be able to run independently. If used, this would typically be
        "<major>.<minor>". Only is applied if app_name is present.
    :param roaming: bool, True to use the Windows roaming appdata directory. That means that for users on a a
        Windows network setup for roaming profiles, this user data will be synced on login. See
        <http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx> for a discussion of issues.
    :return: str

    Typical user data directories are:
        Mac OS X:               ~/Library/Application Support/<AppName>
        Unix:                   ~/.local/share/<AppName>    # or in $XDG_DATA_HOME, if defined
        Win XP (not roaming):   C:\Documents and Settings\<username>\Application Data\<AppAuthor>\<AppName>
        Win XP (roaming):       C:\Documents and Settings\<user>\Local Settings\Application Data\<AppAuthor>\<AppName>
        Win 7  (not roaming):   C:\Users\<username>\AppData\Local\<AppAuthor>\<AppName>
        Win 7  (roaming):       C:\Users\<username>\AppData\Roaming\<AppAuthor>\<AppName>

    For Unix, we follow the XDG spec and support $XDG_DATA_HOME.
    That means, by default "~/.local/share/<AppName>".
    """

    system = osplatform.get_sys_platform()

    if system == "win32":
        if appauthor is None:
            appauthor = appname
        const = roaming and "CSIDL_APPDATA" or "CSIDL_LOCAL_APPDATA"
        path = os.path.normpath(win32.get_win_folder(const))
        if appname:
            if appauthor is not False:
                path = os.path.join(path, appauthor, appname)
            else:
                path = os.path.join(path, appname)
    elif system == 'darwin':
        path = os.path.expanduser('~/Library/Application Support/')
        if appname:
            path = os.path.join(path, appname)
    else:
        path = os.getenv('XDG_DATA_HOME', os.path.expanduser("~/.local/share"))
        if appname:
            path = os.path.join(path, appname)
    if appname and version:
        path = os.path.join(path, version)

    return path



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility methods related to languages
"""

from __future__ import print_function, division, absolute_import

import re
import os
import locale


class Language(object):
    def __init__(self, en='', es='', jp=''):
        self.en = en
        self.es = es
        self.jp = jp

    def output(self):
        lang = 'en'
        env = re.sub('_.+', '', os.environ.get('MAYA_UI_LANGUAGE', ''))
        loc = re.sub('_.+', '', locale.getdefaultlocale()[0])
        env = re.sub('-.+', '', env)
        loc = re.sub('-.+', '', loc)

        if loc != '':
            lang = loc
        if env != '':
            lang = env
        if lang == 'ja' or lang == 'jp':
            return self.jp
        elif lang == 'en':
            return self.en
        elif lang == 'es':
            return self.es
        return self.en



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains Config definition
"""

from __future__ import print_function, division, absolute_import

import os
import logging
import platform
import subprocess
try:
    import configparser
except ImportError:
    import ConfigParser as configparser


from tpDcc.libs.python import osplatform

LOGGER = logging.getLogger('tpDcc-libs-python')


class Config(configparser.RawConfigParser, object):
    """
    Configuration file
    """

    EXCLUDE_PATTERNS = ['__*', '*.']
    ICON_EXTENSIONS = ['xpm', 'png', 'bmp', 'jpeg']

    def __init__(self, app_name, root_folder=None, config_name='config', config_extension='ini', *args, **kwargs):
        super(Config, self).__init__(*args, **kwargs)

        self._app_name = app_name.replace(' ', '_').lower()

        if root_folder:
            self.config_file = os.path.join(
                osplatform.get_system_config_directory(),
                root_folder, self._app_name, '{}.{}'.format(config_name, config_extension))
        else:
            self.config_file = os.path.join(
                osplatform.get_system_config_directory(),
                self._app_name, '{}.{}'.format(config_name, config_extension))

        LOGGER.info('{0}: Configuration File: {1}'.format(self._app_name, self.config_file))
        try:
            self.readfp(open(self.config_file, 'r'))
            LOGGER.info('{0}: Configuration File read successfully!'.format(self._app_name))
        except IOError:
            LOGGER.info('{0}: Configuration file not found! Creating it...'.format(self._app_name))
            self._create()

    @property
    def app_name(self):
        return self._app_name

    @staticmethod
    def create_config(app_name, root_folder=None):
        """
        Construct the configuration object from necessary elements
        """

        config = Config(app_name=app_name, root_folder=root_folder, allow_no_value=True)

        return config

    def _create(self):
        """
        If configuration file is not already created we create it
        """

        LOGGER.info(
            'Initializing {0} Settings, creating configuration file: {1}\n'.format(
                self._app_name, self.config_file))

        self.add_section(self._app_name)

        if not os.path.exists(os.path.dirname(self.config_file)):
            try:
                LOGGER.info('Creating Settings Folder: {}'.format(os.path.dirname(self.config_file)))
                original_umask = os.umask(0)
                os.makedirs(os.path.dirname(self.config_file), 0o770)
            finally:
                os.umask(original_umask)
        f = open(self.config_file, 'w')
        f.close()
        self.update()

        LOGGER.info(
            '{0} has successfully created a new configuration file at: {1}\n'.format(
                self._app_name, str(self.config_file)))

    def update(self):
        with open(self.config_file, 'wb') as f:
            self.write(f)

    def get(self, option, section=None):

        if not section:
            section = self._app_name

        try:
            return configparser.RawConfigParser.get(self, section, option)
        except configparser.NoOptionError:
            return ''

    def get_list(self, option, section=None):
        """
        Convert string value to list object
        """

        if not section:
            section = self._app_name

        if self.has_option(section, option):
            return self.get(section=section, option=option).replace(' ', '').split(',')
        else:
            raise KeyError('{0} with {1} does not exist!'.format(section, option))

    def edit(self):
        """
        Edit file with default OS application
        """

        if platform.system().lower() == 'windows':
            os.startfile(str(self.config_file))
        else:
            if platform.system().lower() == 'darwin':
                call = 'open'
            else:
                call = 'xdg-open'
            subprocess.call(call, self.config_file)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains basic class for Python importers
"""


from __future__ import print_function, division, absolute_import

import os
import sys
import pkgutil
import traceback
import importlib
from collections import OrderedDict

from tpDcc.libs.python import python


def import_module(module_name):
    """
    Static function used to import a function given its complete name
    :param module_name: str, name of the module we want to import
    """

    try:
        mod = importlib.import_module(module_name)
        # print('Imported: {} | {}'.format(module_name, mod))
        return mod
    except Exception:
        try:
            print('FAILED IMPORT: {} -> {}'.format(str(module_name), str(traceback.format_exc())))
        except Exception:
            print('FAILED IMPORT: {}'.format(module_name))


def reload_module(module_to_reload):
    """
    Reloads given module
    :param module_to_reload: mode
    """

    if sys.version[0] <= '2':
        reload(module_to_reload)
    else:
        importlib.reload(module_to_reload)


def import_submodules(package_dot_path, skip_modules, recursive=True):
    """
    Import all the modules of the package
    """

    extra_skip = tuple(['{}.'.format(mod) for mod in skip_modules])

    def _import_submodules(pkg):

        found_modules = list()

        if python.is_string(pkg):
            pkg = import_module(pkg)
        if not pkg:
            return found_modules

        pkg_paths = tuple([os.path.normpath(pkg_path) for pkg_path in pkg.__path__ if pkg is not None])
        for pkg_path in list(pkg_paths):
            for loader, name, is_pkg in pkgutil.walk_packages(pkg_paths):
                loader_path = os.path.normpath(loader.path)
                full_name = pkg.__name__ + '.' + name
                if not loader_path.startswith(pkg_path):
                    # print('Trying to import non valid module: {} | {}'.format(full_name, loader_path))
                    continue
                if full_name in skip_modules or full_name.startswith(extra_skip):
                    # print('Skipping .... {}'.format(full_name))
                    continue
                found_modules.append(full_name)
                if recursive and is_pkg:
                    found_modules.extend(_import_submodules(full_name))

        return found_modules

    modules_to_import = [package_dot_path]
    modules_to_import.extend(list(set(_import_submodules(package_dot_path))))

    loaded_modules = OrderedDict()

    for full_name in modules_to_import:
        loaded_modules[full_name] = import_module(full_name)

    return loaded_modules


class PackageImporter(object):
    """
    Base class that allows to import/reload all the modules in a given package and in a given order
    """

    def __init__(self, package):
        super(PackageImporter, self).__init__()

        self._package = package

        self.loaded_modules = OrderedDict()
        self.reload_modules = list()

    def import_package(self, skip_modules=None):
        skip_modules = skip_modules if skip_modules else list()
        skip_modules = tuple(mod for mod in skip_modules)

        return import_submodules(self._package, skip_modules=skip_modules)


def init_importer(package, skip_modules=None):
    """
    Initializes importer
    :param package:
    :param skip_modules: bool
    :return:
    """

    new_importer = PackageImporter(package)
    new_importer.import_package(skip_modules=skip_modules)

    return new_importer



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains folder related tests for tpDcc-libs-python
"""

from __future__ import print_function, division, absolute_import

import os
import uuid
import shutil
from tpDcc.libs.unittests.core import unittestcase

from tpDcc.libs.python import folder


class TestFolder(unittestcase.UnitTestCase()):
    def __init__(self, *args, **kwargs):
        super(TestFolder, self).__init__(*args, **kwargs)

        self._data_folder = None

    def setUp(self):
        super(TestFolder, self).setUp()

        self._data_folder = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'tests_data')
        if not os.path.isdir(self._data_folder):
            os.makedirs(self._data_folder)

    def tearDown(self):
        super(TestFolder, self).tearDown()

        if os.path.isdir(self._data_folder):
            shutil.rmtree(self._data_folder)

    def _get_folder_path(self, split=False):
        folder_path = os.path.join(self._data_folder, str(uuid.uuid4().hex))
        if split:
            folder_dir = os.path.dirname(folder_path)
            folder_name = os.path.basename(folder_path)
            return folder_dir, folder_name
        else:
            return folder_path


class TestFolderCreation(TestFolder):

    def test_folder_creation_path(self):
        self.assertTrue(os.path.isdir(folder.create_folder(self._get_folder_path())))

    def test_folder_creation_name_directory(self):
        split_folder_path = self._get_folder_path(split=True)
        self.assertTrue(os.path.isdir(folder.create_folder(split_folder_path[1], split_folder_path[0])))

    def test_folder_creation_unique(self):
        folder_path = self._get_folder_path()
        folder.create_folder(folder_path)
        self.assertTrue(os.path.isdir(folder.create_folder(folder_path, make_unique=True)))


class TestFolderRename(TestFolder):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utility functions decorators
"""

from __future__ import print_function, division, absolute_import

import os
import time
import inspect
import logging
import traceback
import threading
from functools import wraps, update_wrapper

from tpDcc.libs.python import python, debug

LOGGER = logging.getLogger('tpDcc-libs-python')


def abstractmethod(fn):
    """
    The decorated function should be overridden by a software specific module.
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        msg = 'Abstract implementation has not been overridden.'
        mode = os.getenv('ABSTRACT_METHOD_MODE')
        if not mode or mode == 'raise':
            raise NotImplementedError(debug.debug_object_string(fn, msg))
        elif mode == 'warn':
            LOGGER.warning(debug.debug_object_string(fn, msg))
        return fn(*args, **kwargs)

    return wrapper


def accepts(*types, **kw):
    """
    Function decorator that checks if decorated function arguments are of the expected types
    :param types: Expected types of the inputs to the decorated function. Must specify a type for each parameter
    :param kw: (optional) Specification of 'debug' level ( 0 | 1 | 2 )
    """

    if not kw:
        is_debug = 1
    else:
        is_debug = kw['debug']
    try:
        def decorator(f):
            def new_fn(*args):
                if debug == 0:
                    return f(*args)
                args = list(args)
                if not (len(args[1:]) == len(types)):
                    raise AssertionError
                arg_types = tuple(map(type, args[1:]))
                if arg_types != types:
                    msg = debug.format_message(f.__name__, types, arg_types, 0)
                    if is_debug == 1:
                        try:
                            for i in range(1, len(args)):
                                args[i] = types[i - 1](args[i])
                        except ValueError:
                            raise ValueError(msg)
                        except TypeError:
                            raise TypeError(msg)
                    elif is_debug == 2:
                        raise TypeError(msg)
                return f(*args)
            new_fn.__name__ = f.__name__
            return new_fn
        return decorator
    except KeyError as key:
        raise KeyError(key + 'is not a valid keyword argument')
    except TypeError as msg:
        raise TypeError(msg)


def returns(ret_type, **kw):
    """
    Function decorator that checks if decorated function return value is of the expected type
    :param retType: Excepted type of the decorated function return value. Must specify type for each parameter
    :param kwargs: (optional) Specification of 'debug' level (0 | 1 | 2)
    """

    try:
        if not kw:
            is_debug = 1
        else:
            is_debug = kw['debug']

        def decorator(f):
            def newf(*args):
                result = f(*args)
                if is_debug == 0:
                    return result
                res_type = type(result)
                if res_type != ret_type:
                    msg = debug.format_message(f.__name__, (ret_type,), (res_type,), 1)
                    if is_debug == 1:
                        try:
                            result = ret_type(result)
                        except ValueError:
                            raise ValueError
                    elif is_debug == 2:
                        raise TypeError(msg)
                return result
            newf.__name__ = f.__name__
            return newf
        return decorator
    except KeyError as key:
        raise KeyError(key + 'is not a valid keyword argument')
    except TypeError as msg:
        raise TypeError(msg)


def timer(fn):
    """
    Function decorator for simple timer
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        if LOGGER.getEffectiveLevel() == 20:
            # If we are in info mode we disable timers
            res = fn(*args, **kwargs)
        else:
            t1 = time.time()
            res = fn(*args, **kwargs)
            t2 = time.time()

            trace = ''
            try:
                mod = inspect.getmodule(args[0])
                trace += '{0} >>'.format(mod.__name__.split('.')[-1])
            except Exception:
                LOGGER.debug('function module inspect failure')

            try:
                trace += '{0}.'.format(args[0].__class__.__name__)
            except Exception:
                LOGGER.debug('function class inspect failure')

            trace += fn.__name__
            LOGGER.debug('Timer : %s: took %0.3f ms' % (trace, (t2 - t1) * 1000.0))
        return res
    return wrapper


def print_elapsed_time(f):
    """
    Function decorator that gets the elapsed time
    :param f: fn, function
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        t = time.time()
        res = f(*args, **kwargs)
        print(f.__name__, time.time() - t)
        return res
    return wrapper


def timestamp(f):
    """
    Function decorator that gets the elapsed time with a more descriptive output
    :param f: fn, function
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        res = f(*args, **kwargs)
        function_name = f.func_name if python.is_python2() else f.__name__
        LOGGER.info('<{}> Elapsed time : {}'.format(function_name, time.time() - start_time))
        return res
    return wrapper


def try_pass(fn):
    """
    Function decorator that tries a function and if it fails pass
    :param fn: function
    """

    def wrapper(*args, **kwargs):
        return_value = None
        try:
            return_value = fn(*args, **kwargs)
        except Exception:
            LOGGER.error(traceback.format_exc())
        return return_value
    return wrapper


if python.is_python2():
    def empty_decorator(*args, **kwargs):
        def fn_decorator(fn):
            def wrapper(*args, **kwargs):
                r = fn(*args, **kwargs)
                return r
            return wrapper
        return fn_decorator
else:
    def empty_decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            r = fn(*args, **kwargs)
            return r
        return wrapper


def repeater(interval, limit=-1):
    """!
    A function interval decorator based on
    http://stackoverflow.com/questions/5179467/equivalent-of-setinterval-in-python

    Infinite Example Usage:
        @repeater(.05)
        def infinite():
            print "Executing infinity."

        import time
        print "Starting Infinite Repeating Function"
        token = infinite()
        time.sleep(1)
        print "Stopping: infinite"
        token.stop()

    Limited Example Usage:
        @repeater(.05, 5)
        def x5():
            print "Executing x5"

        import time
        print "Starting Limited Repeating Function"
        token = x5()
        while not token.isSet():
            time.sleep(1)
        print "Expired: x5"

    @param interval      The interval (in seconds) between function invocations.
    @param limit         The limit to the number of function invocations; -1 represents infinity.
    @return              A new decorator with a closure around the original function and the Python threading.
                         Thread used to invoke it.
    """

    def actual_decorator(fn):

        def wrapper(*args, **kwargs):

            class RepeaterTimerThread(threading.Thread):
                def __init__(self):
                    threading.Thread.__init__(self)
                    self._event = threading.Event()

                def run(self):
                    i = 0
                    while i != limit and not self._event.is_set():
                        self._event.wait(interval)
                        fn(*args, **kwargs)
                        i += 1
                    else:
                        if self._event:
                            self._event.set()

                def stopped(self):
                    return not self._event or self._event.is_set()

                def pause(self):
                    self._event.set()

                def resume(self):
                    self._event.clear()

                def stop(self):
                    self._event.set()
                    self.join()

            token = RepeaterTimerThread()
            token.daemon = True
            token.start()
            return token

        return wrapper

    return actual_decorator


def cached(fn):
    cache = {}

    @wraps(fn)
    def wrapper(*args):
        try:
            return cache[args]
        except KeyError:
            rv = fn(*args)
            cache[args] = rv
            return rv

    return wrapper


def add_method(cls):
    """
    This decorator adds a function to given class instance
    https://medium.com/@mgarod/dynamically-add-a-method-to-a-class-in-python-c49204b85bd6
    :param cls: class instance
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)
        setattr(cls, func.__name__, wrapper)
        # Note we are not binding func, but wrapper which accepts self but does exactly the same as func
        return func     # returning func means func can still be used normally
    return decorator


def add_metaclass(metaclass):
    """
    Decorators that allows to create a class using a metaclass
    https://github.com/benjaminp/six/blob/master/six.py
    :param metaclass:
    :return:
    """

    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        if hasattr(cls, '__qualname__'):
            orig_vars['__qualname__'] = cls.__qualname__
        return metaclass(cls.__name__, cls.__bases__, orig_vars)
    return wrapper


class Singleton(type):

    """
    Singleton decorator as metaclass. Should be used in conjunction with add_metaclass function of this module
    @add_metaclass(Singleton)
    class MyClass(BaseClass, object):
        ...
    """

    def __new__(meta, name, bases, clsdict):
        if any(isinstance(cls, meta) for cls in bases):
            raise TypeError('Cannot inherit from singleton class')
        clsdict['_instance'] = None
        return super(Singleton, meta).__new__(meta, name, bases, clsdict)

    def __call__(cls, *args, **kwargs):
        if not isinstance(cls._instance, cls):
            cls._instance = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instance


class HybridMethod(object):
    """
    Merges a normal method with a classmethod
    First two arguments are (cls, self), where self will match cls if it is a classmethod
    https://stackoverflow.com/a/18078819/2403000
    """

    def __init__(self, fn):
        self._fn = fn

    def __get__(self, obj, cls):
        context = obj if obj is not None else cls

        @wraps(self._fn)
        def hybrid(*args, **kwargs):
            return self._fn(cls, context, *args, **kwargs)

        # Mimic method attributes (not required)
        hybrid.__func__ = hybrid.im_func = self._fn
        hybrid.__self__ = hybrid.im_self = context

        return hybrid


class LazyProperty(property):
    """
    https://github.com/jackmaney/lazy-property/blob/master/lazy_property/__init__.py

    Allow us to initialize an attribute only once. Conerting this:
    class SomeOtherClass(object):
        @property
        def calculation_value(self):
        if not hasattr(self, "_calculation_value"):
            self._calculation_value = do_some_big_calculation()
        return self._calculation_value

    into this:

    class YetAnotherClass(object):
        @lazy_property.LazyProperty
        def calculation_value(self):
            return do_some_big_calculation()
    """

    def __init__(self, method, fget=None, fset=None, fdel=None, doc=None):

        self.method = method
        self.cache_name = "_{}".format(self.method.__name__)

        doc = doc or method.__doc__
        super(LazyProperty, self).__init__(fget=fget, fset=fset, fdel=fdel, doc=doc)

        update_wrapper(self, method)

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if hasattr(instance, self.cache_name):
            result = getattr(instance, self.cache_name)
        else:
            if self.fget is not None:
                result = self.fget(instance)
            else:
                result = self.method(instance)
            setattr(instance, self.cache_name, result)
        return result


class LazyWritableProperty(LazyProperty):
    """
    https://github.com/jackmaney/lazy-property/blob/master/lazy_property/__init__.py
    """

    def __set__(self, instance, value):
        if instance is None:
            raise AttributeError
        if self.fset is None:
            setattr(instance, self.cache_name, value)
        else:
            self.fset(instance, value)
